/***************************************** (C) COPYRIGHT 2010 Intersil *****************************************
 * File Name         : isil_chip.c
 * Author            : tang.sc@intersil.com
 * Revision          : 1.2.1.1275
 * Date              : 2012-03-29
 * Description       : this file is tw5864 driver's root file, driver's insmod&rmmod entrance is here.
 *                   this file focus on all chip level code, include driver's init, pci device's register&init,
 *                   chip isr process, chip reg interface, and so on
 * Change log        : 
 *                   
 ****************************************************************************************************************
 * THE PRESENT SOFTWARE IS DISTRIBUTED IN THE HOPE THAT IT WILL PROVIDE CUSTOMERS 
 * WITH CODING INFORMATION TO SPEED UP THEIR PRODUCT DEVELOPMENT, BUT WITHOUT
 * ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS
 * FOR A PARTICULAR PURPOSE. AS A RESULT, TECHWELL SHALL NOT BE HELD LIABLE FOR ANY
 * DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES WITH RESPECT TO ANY CLAIMS ARISING
 * FROM THE CONTENT OF SUCH SOFTWARE AND/OR THE USE MADE BY CUSTOMERS OF THE
 * CODING INFORMATION CONTAINED HEREIN IN CONNECTION WITH THEIR PRODUCTS.
 ****************************************************************************************************************/

#include <linux/version.h>
#include    <isil5864/isil_common.h>
#include    <isil5864/tc_common.h>

//int mcount = 0;
#if LINUX_VERSION_CODE > KERNEL_VERSION(4,11,0)
#include <linux/sched/signal.h>
#else
#include <linux/sched.h>
#endif

static struct mutex pool_lock;

static	DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY platform_video_encode_cap[1] = {
    {
        .first_device_number = 0,
        .video_encode_work_mode = ENCODE_4_D1,
        .encode_reg_4 = (DAON|DEBLOCKINGON|INTRAON|CMOSON|INTERON|DDRON),
        .fpga_sync_base_addr = 0,
        .fpga_sync_section_len = 0,
        .fpga_async_base_addr = 0x08000000,
        .fpga_async_section_len = 0x100000,
        .ddr_map_mode = DDR_MAP_COMPRESS_DISABLE,
    },
};

#ifdef  POWERPC_PLATFORM
extern void isil_map_init(void);
extern void isil_map_exit(void);
#endif


/****************************************************************************************************************
 * Function Name :   chip_request_irq
 * Description   :   this function provide isr callback register.
 * Input         :   param1 is chip, 
 *                   param2 is isr source number in chip,
 *                   param3 is string for help to see,
 *                   param4 is isr callback's pass param.
 * Output        :   null
 * Return        :   ISIL_OK is meaning register isr service callback ok, other is meaning register err.
 ****************************************************************************************************************/
int chip_request_irq(isil_chip_t *chip, u32 irq, isil_irq_handler_t handler, const char *fnname, void *context)
{
    unsigned long flags;
    struct chip_irq_action *action;
    struct chip_irq_desc   *desc;

    if(!chip || !handler){
        return -EINVAL;
    }

    desc = chip_irq_to_desc(chip, irq);
    if(!desc){
        return -EINVAL;
    }

    action = &desc->action;
    if (!action)
	    return -EINVAL;
    if (action->handler == handler) {
	    printk("duplicate\n");
	    atomic_inc(&desc->depth);
	    //chip_enable_irq(chip, irq);
	    return 0;
    }
    
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 0, 0)
    if(desc->flags & IRQ_DISABLED) {
        return -EFAULT;
    }
#elif LINUX_VERSION_CODE < KERNEL_VERSION(3, 10, 0) // TBD
    if(desc->flags & IRQF_DISABLED) {
        return -EFAULT;
    }
#endif
    spin_lock_irqsave(&desc->chip_irq_desc_lock, flags);
    action = &desc->action;
    action->handler = handler;
    action->irq     = irq;
    action->name    = fnname;
    action->context = context;
    action->next    = NULL;

    desc->name           = fnname;
    //desc->irq_count      = 0;
    //desc->irqs_unhandled = 0;
    atomic_inc(&desc->depth);
    spin_unlock_irqrestore(&desc->chip_irq_desc_lock, flags);

    chip_enable_irq(chip, irq);

    return ISIL_OK;
}

/****************************************************************************************************************
 * Function Name :   chip_free_irq
 * Description   :   this function provide isr callback deregister.
 * Input         :   param1 is chip, 
 *                   param2 is isr source number in chip,
 *                   param3 is isr callback's pass param.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void chip_free_irq(isil_chip_t *chip, u32 irq, void *context)
{
    struct chip_irq_desc    *desc;
    struct chip_irq_action  *action;
    unsigned long flags;

    desc = chip_irq_to_desc(chip, irq);
    if(!desc){
        return;
    }
    spin_lock_irqsave(&desc->chip_irq_desc_lock, flags);
    action = &desc->action;

    action->handler = NULL;
    action->irq     = irq;
    action->name    = NULL;
    action->context = NULL;
    action->next    = NULL;
    atomic_dec(&desc->depth);

    spin_unlock_irqrestore(&desc->chip_irq_desc_lock, flags);

    chip_disable_irq(chip, irq);

    return ;
}

/****************************************************************************************************************
 * Function Name :   chip_enable_irq
 * Description   :   this function is enable process this number source of isr.
 * Input         :   param1 is chip, 
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void chip_enable_irq(isil_chip_t *chip, u32 irq)
{
    isil_interrupt_control_t *irq_ctrl;

    if(chip && (irq < CHIP_IRQS_NR)){
        irq_ctrl = &chip->isil_int_control;
        irq_ctrl->system_op->enable_interrupt_source(irq_ctrl, irq);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_disable_irq
 * Description   :   this function is disable process this number source of isr.
 * Input         :   param1 is chip, 
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void chip_disable_irq(isil_chip_t *chip, u32 irq)
{
    isil_interrupt_control_t *irq_ctrl;

    if(chip && (irq < CHIP_IRQS_NR)){
        irq_ctrl = &chip->isil_int_control;
        irq_ctrl->system_op->disable_interrupt_source(irq_ctrl, irq);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_clear_irq
 * Description   :   this function is clear process this number source of isr.
 * Input         :   param1 is chip, 
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void chip_clear_irq(isil_chip_t *chip, u32 irq)
{
    isil_interrupt_control_t *irq_ctrl;

    if(chip && (irq < CHIP_IRQS_NR)){
        irq_ctrl = &chip->isil_int_control;
        irq_ctrl->hardware_op->ack_interrupt(irq_ctrl, irq);
    }
}

/****************************************************************************************************************
 * Function Name :   isil_vlc_irq_handler
 * Description   :   vlc interupt response function.
 * Input         :   param1 is isr source number in chip.
 *                   param2 is passing param.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static isil_irq_handler_t isil_vlc_irq_handler(int irq, void *context)
{
    isil_chip_t	*chip;

    chip = context;
    if(chip != NULL){
        if(chip->curr_h264_encode_chan != NULL) {
            chip->curr_h264_encode_chan->encode_control.op->irq_func(irq, (void*)&chip->curr_h264_encode_chan->encode_control);
        } else {
            ISIL_DBG(ISIL_DBG_INFO, ": no register isr curr service chan\n");
        }
    }

    return 0;
}

#if defined(HISILICON_PLATFORM)
#include <asm/arch/pci.h>
#endif
/****************************************************************************************************************
 * Function Name :   isil_top_isr
 * Description   :   chip isr's root entrance.
 * Input         :   param1 is cpu's isr number
 *                   param2 is system deliver to us param.
 * Output        :   null
 * Return        :   irqreturn_t
 ****************************************************************************************************************/
static irqreturn_t  isil_top_isr(int irq, void *id)
{
    isil_interrupt_control_t *control = (isil_interrupt_control_t*)id;
    isil_chip_t	*chip = control->chip;
    u32     chip_irq = 0;
    struct chip_irq_desc *desc;
    struct chip_irq_action *action;

#if defined(HISILICON_PLATFORM)
    hi3520_bridge_ahb_writel(CPU_ISTATUS, 0x1000000);
#endif

    control->hardware_op->get_interrupt_status(control);
    if(control->chip_interrupt_status){
        for(chip_irq = 0; chip_irq < CHIP_IRQS_NR; chip_irq++){
            if(control->chip_interrupt_status & (1 << chip_irq)){
                desc = chip_irq_to_desc(chip, chip_irq);
                desc->irq_count++;
                action = &desc->action;
                if(action->handler) {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 18)
                    if(action->handler(action->irq, action->context, NULL)){
                        /*nothing*/
                    }
#else
                    if(action->handler(action->irq, action->context)){
                        /*nothing*/
                    }
#endif
                }else{
                    desc->irqs_unhandled++;
                }
                chip_clear_irq(chip, chip_irq);
            }
        }
        return IRQ_HANDLED;
    } else {
        return IRQ_NONE;
    }
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_init
 * Description   :   5864 chip interrupt controller's init function.
 * Input         :   param1 is 5864's interrupt controller management struct
 *                   param2 is 5864's chip management struct
 * Output        :   null
 * Return        :   ISIL_ERR meaning init interrupt controller err.
 *                   ISIL_OK meaning init interrupt controller ok.
 ****************************************************************************************************************/
static int  isil_interrupt_control_system_interface_init(isil_interrupt_control_t *control, isil_chip_t *chip)
{
    int	ret = ISIL_ERR;
    u32 i;

    if(control != NULL){
        spin_lock_init(&control->isil_interrupt_control_lock);
        control->interrupt_source = 0x00000000;
        control->interrupt_source_trigger_mode = 0x00000000;
        control->interrupt_source_assert = 0xffffffff;
        control->chip_interrupt_output_assert = 0x0;
        control->chip_interrupt_status = 0x00000000;
        control->chip_total_duration = 0;
        control->chip = chip;
        /*level mode*/
        chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_TRIGGER_REG_L, (control->interrupt_source_trigger_mode & 0xffff));
        chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_TRIGGER_REG_H, ((control->interrupt_source_trigger_mode>>16) & 0xffff));

        chip->io_op->chip_write32(chip, INTERRUPT_ASSERT_REG_L, (control->interrupt_source_assert & 0xffff));
        chip->io_op->chip_write32(chip, INTERRUPT_ASSERT_REG_H, ((control->interrupt_source_assert>>16) & 0xffff));

        chip->io_op->chip_write32(chip, INTERRUPT_OUTPUT_ASSERT_REG, control->chip_interrupt_output_assert);

        /*init irq describe table*/
        for(i = 0; i < CHIP_IRQS_NR; i++)
        {
            struct chip_irq_desc *desc = chip_irq_to_desc(chip, i);
            desc->irq = i;
            desc->irq_count =0;
            desc->irqs_unhandled = 0;
            desc->irq_total_duration = 0;
            spin_lock_init(&desc->chip_irq_desc_lock);
            atomic_set(&desc->depth, 0);
            memset(&desc->action, 0, sizeof(struct chip_irq_action));
            switch(desc->irq)
            {
                case IRQ_VLC_TYPE_INTR:
                    desc->name = "VLC";break;
                case IRQ_BURST_TYPE_INTR:
                    desc->name = "BURST";break;
                case IRQ_MV_TYPE_INTR:
                    desc->name = "MV";break;
                case IRQ_FRONT_END_TYPE_INTR:
                    desc->name = "FRONT_END";break;
                case IRQ_JPEG_TYPE_INTR:
                    desc->name = "JPEG";break;
                case IRQ_TIMER_TYPE_INTR:
                    desc->name = "TIMER";break;
                case IRQ_VLC_DONE_TYPE_INTR:
                    desc->name = "VLC_DONE";break;
                case IRQ_AD_SYNC_TYPE_INTR:
                    desc->name = "AD_SYNC";break;
                case IRQ_AUDIO_EOF_TYPE_INTR:
                    desc->name = "AUDIO";break;
                case IRQ_IIC_DONE_INTR:
                    desc->name = "IIC";break;
                case IRQ_AD_EVENT_INTR:
                    desc->name = "AD_EVENT";break;
                default:
                    desc->name = "UNDEFINE";
            }
            chip_disable_irq(chip, i);
        }
        ret = ISIL_OK;
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_enable_interrupt_source
 * Description   :   enable 5864 chip's some interrupt source.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_interrupt_control_system_interface_enable_interrupt_source(isil_interrupt_control_t *control, u32 source)
{
    if((control != NULL) && (source < CHIP_IRQS_NR)){
        isil_chip_t  *chip;
        unsigned long	flags;
        spin_lock_irqsave(&control->isil_interrupt_control_lock, flags);
        chip = control->chip;

        control->interrupt_source &= ~(INTERRUPT_DISABLE<<source);
        control->interrupt_source |= (INTERRUPT_ENABLE<<source);
        if(source < (CHIP_IRQS_NR / 2)){
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_ENABLE_REG_L, (control->interrupt_source & 0xffff));
        } else{
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_ENABLE_REG_H, ((control->interrupt_source >> 16) & 0xffff));
        }
        spin_unlock_irqrestore(&control->isil_interrupt_control_lock, flags);
    }
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_disable_interrupt_source
 * Description   :   disable 5864 chip's some interrupt source.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_interrupt_control_system_interface_disable_interrupt_source(isil_interrupt_control_t *control, u32 source)
{
    if((control != NULL) && (source < CHIP_IRQS_NR)){
        isil_chip_t  *chip;
        unsigned long	flags;
        spin_lock_irqsave(&control->isil_interrupt_control_lock, flags);
        chip = control->chip;
        control->interrupt_source &= ~(INTERRUPT_ENABLE<<source);
        control->interrupt_source |= (INTERRUPT_DISABLE<<source);
        if(source < (CHIP_IRQS_NR / 2)){
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_ENABLE_REG_L, (control->interrupt_source & 0xffff));
        } else {
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_ENABLE_REG_H, ((control->interrupt_source >> 16) & 0xffff));
        }
        spin_unlock_irqrestore(&control->isil_interrupt_control_lock, flags);
    }
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_set_interrupt_source_edge_trigger
 * Description   :   set interrupt source trigger controller by edge mode.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_interrupt_control_system_interface_set_interrupt_source_edge_trigger(isil_interrupt_control_t *control, u32 source)
{
    if((control != NULL) && (source < CHIP_IRQS_NR)){
        isil_chip_t  *chip;
        unsigned long	flags;
        spin_lock_irqsave(&control->isil_interrupt_control_lock, flags);
        chip = control->chip;
        control->interrupt_source_trigger_mode &= ~(LEVEL_TRIGGER<<source);
        control->interrupt_source_trigger_mode |= (EDGE_TRIGGER<<source);
        if(source < (CHIP_IRQS_NR / 2)){
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_TRIGGER_REG_L, (control->interrupt_source_trigger_mode & 0xffff));
        } else{
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_TRIGGER_REG_H, ((control->interrupt_source_trigger_mode >> 16) & 0xffff));
        }
        spin_unlock_irqrestore(&control->isil_interrupt_control_lock, flags);
    }
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_set_interrupt_source_level_trigger
 * Description   :   set interrupt source trigger controller by level mode.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_interrupt_control_system_interface_set_interrupt_source_level_trigger(isil_interrupt_control_t *control, u32 source)
{
    if((control != NULL) && (source < CHIP_IRQS_NR)){
        isil_chip_t  *chip;
        unsigned long	flags;
        spin_lock_irqsave(&control->isil_interrupt_control_lock, flags);
        chip = control->chip;
        control->interrupt_source_trigger_mode &= ~(EDGE_TRIGGER<<source);
        control->interrupt_source_trigger_mode |= (LEVEL_TRIGGER<<source);
        if(source < (CHIP_IRQS_NR / 2)){
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_TRIGGER_REG_L, (control->interrupt_source_trigger_mode & 0xffff));
        } else{
            chip->io_op->chip_write32(chip, INTERRUPT_SOURCE_TRIGGER_REG_H, ((control->interrupt_source_trigger_mode >> 16) & 0xffff));
        }
        spin_unlock_irqrestore(&control->isil_interrupt_control_lock, flags);
    }
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_calc_irq_ratio_with_system
 * Description   :   this function is help function parse isr profiling.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_interrupt_control_system_interface_calc_irq_ratio_with_system(isil_interrupt_control_t *control, isil_chip_t *chip)
{
    struct chip_irq_desc *desc;
    struct chip_irq_desc *timer_desc;
    u64 total_irq_duration;
    u64 total_system_duration=0;
    int i;

    total_irq_duration = 0;
    for(i = 0; i < CHIP_IRQS_NR; i++){
        desc = chip_irq_to_desc(chip, i);
        total_irq_duration += desc->irq_total_duration;
    }
    timer_desc = chip_irq_to_desc(chip, IRQ_TIMER_TYPE_INTR);
    control->chip_total_duration = timer_desc->irq_count*MS_PER_TICK;
    total_system_duration = control->chip_total_duration >> 10;     //change ms to s
    total_irq_duration = total_irq_duration>>10;                    //change us to ms
    if(total_system_duration && total_irq_duration){
        control->irq_ratio_with_system = ((u32)total_irq_duration)/((u32)total_system_duration);
    } else {
        control->irq_ratio_with_system = 0;
    }
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_get_irq_ratio
 * Description   :   this function is help function parse isr profiling.
 * Input         :   param1 is 5864's interrupt controller management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static u32  isil_interrupt_control_system_interface_get_irq_ratio(isil_interrupt_control_t *control)
{
    return control->irq_ratio_with_system;
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_system_interface_get_system_duration
 * Description   :   this function is help function parse isr profiling.
 * Input         :   param1 is 5864's interrupt controller management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static u64  isil_interrupt_control_system_interface_get_system_duration(isil_interrupt_control_t *control)
{
    return control->chip_total_duration;
}

static struct isil_interrupt_control_system_interface_operation	isil_interrupt_control_system_interface_op = {
    .init = isil_interrupt_control_system_interface_init,
    .enable_interrupt_source = isil_interrupt_control_system_interface_enable_interrupt_source,
    .disable_interrupt_source = isil_interrupt_control_system_interface_disable_interrupt_source,
    .set_interrupt_source_edge_trigger = isil_interrupt_control_system_interface_set_interrupt_source_edge_trigger,
    .set_interrupt_source_level_trigger = isil_interrupt_control_system_interface_set_interrupt_source_level_trigger,
    .calc_irq_ratio_with_system = isil_interrupt_control_system_interface_calc_irq_ratio_with_system,
    .get_irq_ratio = isil_interrupt_control_system_interface_get_irq_ratio,
    .get_system_duration = isil_interrupt_control_system_interface_get_system_duration,
};

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_hardware_interface_get_interrupt_status
 * Description   :   this function get chip interrupt pending status.
 * Input         :   param1 is 5864's interrupt controller management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_interrupt_control_hardware_interface_get_interrupt_status(isil_interrupt_control_t *control)
{
    if(control != NULL){
        isil_chip_t  *chip;
        unsigned long	flags;
        spin_lock_irqsave(&control->isil_interrupt_control_lock, flags);
        chip = control->chip;
        control->chip_interrupt_status = (chip->io_op->chip_read32(chip, INTERRUPT_STATUS_REG_L)) & 0xffff;
        control->chip_interrupt_status |= ((chip->io_op->chip_read32(chip, INTERRUPT_STATUS_REG_H) & 0xffff) << 16);
        spin_unlock_irqrestore(&control->isil_interrupt_control_lock, flags);
    }
}

/****************************************************************************************************************
 * Function Name :   isil_interrupt_control_hardware_interface_ack_interrupt
 * Description   :   this function is clear chip interrupt pending status.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is isr source number in chip.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void isil_interrupt_control_hardware_interface_ack_interrupt(isil_interrupt_control_t *control, u32 source)
{
    if((control != NULL) && (source < CHIP_IRQS_NR)){
        isil_chip_t  *chip;
        unsigned long	flags;
        spin_lock_irqsave(&control->isil_interrupt_control_lock, flags);
        chip = control->chip;
        control->chip_interrupt_status &= ~(1<<source);
        if(source < (CHIP_IRQS_NR / 2)){
            chip->io_op->chip_write32(chip, INTERRUPT_CLEAR_REG_L, (1<<source));
        }else{
            chip->io_op->chip_write32(chip, INTERRUPT_CLEAR_REG_H, (1<<(source - 16)));
        }
        spin_unlock_irqrestore(&control->isil_interrupt_control_lock, flags);
    }
}

struct isil_interrupt_control_hardware_interface_operation    isil_interrupt_control_hardware_interface_op = {
    .get_interrupt_status = isil_interrupt_control_hardware_interface_get_interrupt_status,
    .ack_interrupt = isil_interrupt_control_hardware_interface_ack_interrupt,
};

/****************************************************************************************************************
 * Function Name :   init_isil_interrupt_control
 * Description   :   this function is init chip interrupt controller.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int  init_isil_interrupt_control(isil_interrupt_control_t *control, isil_chip_t *chip)
{
    int	ret = ISIL_ERR;

    if((control!=NULL) && (chip!=NULL)){
        control->system_op = &isil_interrupt_control_system_interface_op;
        control->hardware_op = &isil_interrupt_control_hardware_interface_op;
        ret = control->system_op->init(control, chip);
        if(atomic_read(&control->chip_irq_register_flags) == 0){
            atomic_set(&control->chip_irq_register_flags, 1);
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 0, 0)
            set_irq_type(control->chip_irq, IRQ_TYPE_LEVEL_LOW);
#else
            irq_set_irq_type(control->chip_irq, IRQ_TYPE_LEVEL_LOW);
#endif
            ret = request_irq(control->chip_irq, isil_top_isr, IRQF_SHARED, chip->name, control);
            if(ret) {
                atomic_set(&control->chip_irq_register_flags, 0);
                ret = ISIL_ERR;
                ISIL_DBG(ISIL_DBG_INFO, ": install irq failed!, %d\n", ret);
            } else {
                init_isil_chip_kthread_interface(chip);
                start_isil_chip_kthread_interface(chip);
#ifdef	ARM_PLATFORM
                s3c2410_gpio_pullup(S3C2410_GPF1, 0);
                s3c2410_gpio_cfgpin(S3C2410_GPF1, S3C2410_GPF1_EINT1);
#endif
                ret = ISIL_OK;
            }
        }
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   remove_isil_interrupt_control
 * Description   :   this function is remove chip interrupt controller.
 * Input         :   param1 is 5864's interrupt controller management struct.
 *                   param2 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void remove_isil_interrupt_control(isil_interrupt_control_t *control, isil_chip_t *chip)
{
    if((control!=NULL) && (chip!=NULL)){
        stop_isil_chip_kthread_interface(chip);
        if(atomic_read(&control->chip_irq_register_flags)){
            atomic_set(&control->chip_irq_register_flags, 0);
            free_irq(control->chip_irq, control);
        }
        remove_isil_chip_kthread_interface(chip);
        control->system_op->init(control, chip);
    }
}

/****************************************************************************************************************
 * Function Name :   disable_chip_int
 * Description   :   mask chip interrupt response.
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    disable_chip_int(isil_chip_t *chip)
{
    disable_irq(chip->isil_int_control.chip_irq);
}

/****************************************************************************************************************
 * Function Name :   enable_chip_int
 * Description   :   enable chip interrupt response.
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    enable_chip_int(isil_chip_t *chip)
{
    enable_irq(chip->isil_int_control.chip_irq);
}

/************************************************************************************/
/************************************************************************************/
static const fpga_reg_addr  ForwardQuantizationTable[QUANTIZATION_TABLE_LEN] =
{
    0x3333, 0x1f82, 0x3333, 0x1F82, 0x1F82, 0x147B, 0x1F82, 0x147B, 0x3333, 0x1F82, 
    0x3333, 0x1F82, 0x1F82, 0x147B, 0x1F82, 0x147B, 0x2E8C, 0x1D42, 0x2E8C, 0x1D42,
    0x1D42, 0x1234, 0x1D42, 0x1234, 0x2E8C, 0x1D42, 0x2E8C, 0x1D42, 0x1D42, 0x1234,
    0x1D42, 0x1234, 0x2762, 0x199A, 0x2762, 0x199A, 0x199A, 0x1062, 0x199A, 0x1062,
    0x2762, 0x199A, 0x2762, 0x199A, 0x199A, 0x1062, 0x199A, 0x1062, 0x2492, 0x16C1,
    0x2492, 0x16C1, 0x16C1, 0x0E3F, 0x16C1, 0x0E3F, 0x2492, 0x16C1, 0x2492, 0x16C1,
    0x16C1, 0x0E3F, 0x16C1, 0x0E3F, 0x2000, 0x147B, 0x2000, 0x147B, 0x147B, 0x0D1B,
    0x147B, 0x0D1B, 0x2000, 0x147B, 0x2000, 0x147B, 0x147B, 0x0D1B, 0x147B, 0x0D1B,
    0x1C72, 0x11CF, 0x1C72, 0x11CF, 0x11CF, 0x0B4D, 0x11CF, 0x0B4D, 0x1C72, 0x11CF,
    0x1C72, 0x11CF, 0x11CF, 0x0B4D, 0x11CF, 0x0B4D
};

static const fpga_reg_addr  InverseQuantizationTable[QUANTIZATION_TABLE_LEN] =
{
    0x800A, 0x800D, 0x800A, 0x800D, 0x800D, 0x8010, 0x800D, 0x8010, 0x800A, 0x800D, 
    0x800A, 0x800D, 0x800D, 0x8010, 0x800D, 0x8010, 0x800B, 0x800E, 0x800B, 0x800E,
    0x800E, 0x8012, 0x800E, 0x8012, 0x800B, 0x800E, 0x800B, 0x800E, 0x800E, 0x8012,
    0x800E, 0x8012, 0x800D, 0x8010, 0x800D, 0x8010, 0x8010, 0x8014, 0x8010, 0x8014,
    0x800D, 0x8010, 0x800D, 0x8010, 0x8010, 0x8014, 0x8010, 0x8014, 0x800E, 0x8012,
    0x800E, 0x8012, 0x8012, 0x8017, 0x8012, 0x8017, 0x800E, 0x8012, 0x800E, 0x8012,
    0x8012, 0x8017, 0x8012, 0x8017, 0x8010, 0x8014, 0x8010, 0x8014, 0x8014, 0x8019,
    0x8014, 0x8019, 0x8010, 0x8014, 0x8010, 0x8014, 0x8014, 0x8019, 0x8014, 0x8019,
    0x8012, 0x8017, 0x8012, 0x8017, 0x8017, 0x801D, 0x8017, 0x801D, 0x8012, 0x8017,
    0x8012, 0x8017, 0x8017, 0x801D, 0x8017, 0x801D
};

static const fpga_reg_addr	encoder_vlc_lookup_table[VLC_LOOKUP_TABLE_LEN] = {
    0x011,0x000,0x000,0x000,0x065,0x021,0x000,0x000,0x087,0x064,0x031,0x000,
    0x097,0x086,0x075,0x053,0x0a7,0x096,0x085,0x063,0x0b7,0x0a6,0x095,0x074,
    0x0df,0x0b6,0x0a5,0x084,0x0db,0x0de,0x0b5,0x094,0x0d8,0x0da,0x0dd,0x0a4,
    0x0ef,0x0ee,0x0d9,0x0b4,0x0eb,0x0ea,0x0ed,0x0dc,0x0ff,0x0fe,0x0e9,0x0ec,
    0x0fb,0x0fa,0x0fd,0x0e8,0x10f,0x0f1,0x0f9,0x0fc,0x10b,0x10e,0x10d,0x0f8,
    0x107,0x10a,0x109,0x10c,0x104,0x106,0x105,0x108,0x023,0x000,0x000,0x000,
    0x06b,0x022,0x000,0x000,0x067,0x057,0x033,0x000,0x077,0x06a,0x069,0x045,
    0x087,0x066,0x065,0x044,0x084,0x076,0x075,0x056,0x097,0x086,0x085,0x068,
    0x0bf,0x096,0x095,0x064,0x0bb,0x0be,0x0bd,0x074,0x0cf,0x0ba,0x0b9,0x094,
    0x0cb,0x0ce,0x0cd,0x0bc,0x0c8,0x0ca,0x0c9,0x0b8,0x0df,0x0de,0x0dd,0x0cc,
    0x0db,0x0da,0x0d9,0x0dc,0x0d7,0x0eb,0x0d6,0x0d8,0x0e9,0x0e8,0x0ea,0x0d1,
    0x0e7,0x0e6,0x0e5,0x0e4,0x04f,0x000,0x000,0x000,0x06f,0x04e,0x000,0x000,
    0x06b,0x05f,0x04d,0x000,0x068,0x05c,0x05e,0x04c,0x07f,0x05a,0x05b,0x04b,
    0x07b,0x058,0x059,0x04a,0x079,0x06e,0x06d,0x049,0x078,0x06a,0x069,0x048,
    0x08f,0x07e,0x07d,0x05d,0x08b,0x08e,0x07a,0x06c,0x09f,0x08a,0x08d,0x07c,
    0x09b,0x09e,0x089,0x08c,0x098,0x09a,0x09d,0x088,0x0ad,0x097,0x099,0x09c,
    0x0a9,0x0ac,0x0ab,0x0aa,0x0a5,0x0a8,0x0a7,0x0a6,0x0a1,0x0a4,0x0a3,0x0a2,
    0x021,0x000,0x000,0x000,0x067,0x011,0x000,0x000,0x064,0x066,0x031,0x000,
    0x063,0x073,0x072,0x065,0x062,0x083,0x082,0x070,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x011,0x010,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x011,0x021,0x020,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x023,0x022,0x021,0x020,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x023,0x022,0x021,0x031,0x030,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x023,0x022,0x033,0x032,0x031,0x030,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x023,0x030,0x031,0x033,
    0x032,0x035,0x034,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x037,0x036,0x035,0x034,0x033,0x032,0x031,0x041,0x051,0x061,0x071,0x081,
    0x091,0x0a1,0x0b1,0x000,0x002,0x000,0x0e4,0x011,0x0f4,0x002,0x024,0x003,
    0x005,0x012,0x034,0x013,0x065,0x024,0x013,0x063,0x015,0x022,0x075,0x034,
    0x044,0x023,0x023,0x073,0x054,0x033,0x033,0x004,0x043,0x014,0x011,0x043,
    0x014,0x001,0x025,0x015,0x035,0x025,0x064,0x055,0x045,0x035,0x074,0x065,
    0x085,0x0d5,0x012,0x095,0x055,0x045,0x095,0x0e5,0x084,0x075,0x022,0x0a5,
    0x094,0x085,0x032,0x0b5,0x003,0x0c5,0x001,0x044,0x0a5,0x032,0x0b5,0x094,
    0x0c5,0x0a4,0x0a4,0x054,0x0d5,0x0b4,0x0b4,0x064,0x0f5,0x0f5,0x053,0x0d4,
    0x0e5,0x0c4,0x105,0x105,0x0c4,0x074,0x063,0x0e4,0x0d4,0x084,0x073,0x0f4,
    0x004,0x005,0x000,0x053,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x011,0x021,0x031,0x030,0x011,0x021,0x020,0x000,
    0x011,0x010,0x000,0x000,0x011,0x033,0x032,0x043,0x042,0x053,0x052,0x063,
    0x062,0x073,0x072,0x083,0x082,0x093,0x092,0x091,0x037,0x036,0x035,0x034,
    0x033,0x045,0x044,0x043,0x042,0x053,0x052,0x063,0x062,0x061,0x060,0x000,
    0x045,0x037,0x036,0x035,0x044,0x043,0x034,0x033,0x042,0x053,0x052,0x061,
    0x051,0x060,0x000,0x000,0x053,0x037,0x045,0x044,0x036,0x035,0x034,0x043,
    0x033,0x042,0x052,0x051,0x050,0x000,0x000,0x000,0x045,0x044,0x043,0x037,
    0x036,0x035,0x034,0x033,0x042,0x051,0x041,0x050,0x000,0x000,0x000,0x000,
    0x061,0x051,0x037,0x036,0x035,0x034,0x033,0x032,0x041,0x031,0x060,0x000,
    0x000,0x000,0x000,0x000,0x061,0x051,0x035,0x034,0x033,0x023,0x032,0x041,
    0x031,0x060,0x000,0x000,0x000,0x000,0x000,0x000,0x061,0x041,0x051,0x033,
    0x023,0x022,0x032,0x031,0x060,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x061,0x060,0x041,0x023,0x022,0x031,0x021,0x051,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x051,0x050,0x031,0x023,0x022,0x021,0x041,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x040,0x041,0x031,0x032,
    0x011,0x033,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x040,0x041,0x021,0x011,0x031,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x030,0x031,0x011,0x021,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x020,0x021,0x011,0x000,
    0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x010,0x011,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
    0x000,0x000,0x000,0x000
};

/****************************************************************************************************************
 * Function Name :   WriteEncodeVLCLookupTable
 * Description   :   write vlc encode's lookup table
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void WriteEncodeVLCLookupTable(isil_chip_t *chip)
{
    int i;

    chip->io_op->chip_write32(chip, ENABLE_VLC_LOOKUP_TABLE, 0x1);
    for(i=0; i<VLC_LOOKUP_TABLE_LEN; i++){
        chip->io_op->chip_write32(chip, (ENCODE_VLC_LOOKUP_TABLE_BASE+(i<<2)), encoder_vlc_lookup_table[i]);
    }
    chip->io_op->chip_write32(chip, ENABLE_VLC_LOOKUP_TABLE, 0x0);
}

/****************************************************************************************************************
 * Function Name :   WriteForwardQuantizationTable
 * Description   :   write forward quantization table
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void WriteForwardQuantizationTable(isil_chip_t *chip)
{
    int i;

    for (i=0; i<QUANTIZATION_TABLE_LEN; i++) {
        chip->io_op->chip_write32(chip, (QUANTIZATION+(i<<2)), ForwardQuantizationTable[i]);
    }
}

/****************************************************************************************************************
 * Function Name :   WriteInverseQuantizationTable
 * Description   :   write inverse quantization table
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void WriteInverseQuantizationTable(isil_chip_t *chip)
{
    int i;

    for (i=0; i<QUANTIZATION_TABLE_LEN; i++) {
        chip->io_op->chip_write32(chip, (QUANTIZATION+(i<<2)), InverseQuantizationTable[i]);
    }
}

/****************************************************************************************************************
 * Function Name :   start_chip_robust_process
 * Description   :   start chip's robust flow
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    start_chip_robust_process(isil_chip_t *chip)
{
    isil_kthread_msg_queue_t    *isil_msg_queue;
    robust_processing_control_t *chip_robust;

    chip_robust = &chip->chip_robust_processing;
    chip_robust->op->start_robust_processing(chip_robust);
    get_msg_queue_header(&chip->chip_kthread_interface, &isil_msg_queue);
    delete_all_msg_of_this_chip(isil_msg_queue, chip);
}

/****************************************************************************************************************
 * Function Name :   chip_robust_process_done
 * Description   :   notify chip's robust process is over
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    chip_robust_process_done(isil_chip_t *chip)
{
    robust_processing_control_t *chip_robust;

    chip_robust = &chip->chip_robust_processing;
    chip_robust->op->robust_processing_done(chip_robust);
}

/****************************************************************************************************************
 * Function Name :   chip_wait_robust_process_done
 * Description   :   wait chip's robust process over
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
void    chip_wait_robust_process_done(isil_chip_t *chip)
{
    robust_processing_control_t *chip_robust_processing;
    chip_robust_processing = &chip->chip_robust_processing;
    chip_robust_processing->op->wait_robust_processing_done(chip_robust_processing);
}

/****************************************************************************************************************
 * Function Name :   chip_is_in_robust_processing
 * Description   :   check chip robust's status
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
int chip_is_in_robust_processing(isil_chip_t *chip)
{
    robust_processing_control_t *chip_robust_processing;
    chip_robust_processing = &chip->chip_robust_processing;
    return chip_robust_processing->op->is_in_robust_processing(chip_robust_processing);
}

/****************************************************************************************************************
 * Function Name :   timer_on_chip_handler
 * Description   :   chip's hardware timer interrupt response service
 * Input         :   param1 is interrupt source number.
 *                   param2 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int  timer_on_chip_handler(int irq, void *context)
{
    if(context) {
        u32 flags;
        isil_chip_t *chip = (isil_chip_t *)context;
        isil_chip_timer_controller_t    *chip_timer_cont;

        flags = chip->io_op->chip_read32(chip, INTERRUPT_FLAGS_EXT);
        if(flags & (1 << IRQ_EXT_TIMER_INTR)) {
            chip_timer_cont = &chip->chip_timer_cont;
            chip_timer_cont->op->isr(chip_timer_cont);
            /*clear timer pending*/
            chip->io_op->chip_write32(chip, INTERRUPT_FLAGS_EXT, 1 << IRQ_EXT_TIMER_INTR);
        }
    }

    return 1;
}

#if defined(PCI_PLATFORM)

#else
        extern void	closeS3c2410Timer3(void);
        extern void	enableS3c2410Timer3(void);
static irqreturn_t  s3c2410_timer3_isr(int irq, void* id)
{
    if(id != NULL){
        isil_chip_t *chip = (isil_chip_t *)id;
        isil_chip_timer_controller_t    *chip_timer_cont;
        unsigned long   flags;

        local_irq_save(flags);
        chip_timer_cont = &chip->chip_timer_cont;
        chip_timer_cont->op->isr(chip_timer_cont);
        local_irq_restore(flags);
    }
    return IRQ_HANDLED;
}
#endif

/****************************************************************************************************************
 * Function Name :   register_on_chip_timer
 * Description   :   register chip's hardware timer interrupt response service
 * Input         :   param1 is interrupt source number.
 *                   param2 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void register_on_chip_timer(isil_chip_t *chip, isil_irq_handler_t handler)
{
    if(chip && handler) {
#if	defined(PCI_PLATFORM)
        if(chip_request_irq(chip, IRQ_TIMER_TYPE_INTR, timer_on_chip_handler, "timer", chip)){
            ISIL_DBG(ISIL_DBG_INFO, ": register system timer failed\n");
            return;
        }
        chip->io_op->chip_write32(chip, PCI_TIMER_CONTROL, 0x2);/*default is 0->1ms, 1->2ms, 2->4ms, 3->8ms*/
        chip->enable_timer(chip);
#else
        if(request_irq(IRQ_TIMER3, s3c2410_timer3_isr, IRQF_SHARED, "s3c2410_timer3", chip)) {
            closeS3c2410Timer3();
        } else {
            enableS3c2410Timer3();
        }
#endif
    }
}

/****************************************************************************************************************
 * Function Name :   enable_on_chip_timer
 * Description   :   enable chip's hardware timer interrupt
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void enable_on_chip_timer(isil_chip_t *chip)
{
    if(chip) {
        u32 val;

        val = chip->io_op->chip_read32(chip, PCI_MASTER_CONTROL);
        val |= (1 << IRQ_EXT_TIMER_INTR);
        chip->io_op->chip_write32(chip, PCI_MASTER_CONTROL, val);
    }
}

/****************************************************************************************************************
 * Function Name :   disable_on_chip_timer
 * Description   :   disable chip's hardware timer interrupt
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void disable_on_chip_timer(isil_chip_t *chip)
{
    if(chip) {
        u32 val;

        val = chip->io_op->chip_read32(chip, PCI_MASTER_CONTROL);
        val &= ~(1 << IRQ_EXT_TIMER_INTR);
        chip->io_op->chip_write32(chip, PCI_MASTER_CONTROL, val);
    }
}

/****************************************************************************************************************
 * Function Name :   unregister_on_chip_timer
 * Description   :   unregister chip's hardware timer interrupt response service
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void unregister_on_chip_timer(isil_chip_t *chip)
{
    if(chip) {
#if	defined(PCI_PLATFORM)
        chip->disable_timer(chip);
        chip_free_irq(chip, IRQ_TIMER_TYPE_INTR, chip);
#else
        closeS3c2410Timer3();
        free_irq(IRQ_TIMER3, chip);
#endif
    }
}

/****************************************************************************************************************
 * Function Name :   chip_hardware_resource_open
 * Description   :   chip's hardware resource counter
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int  chip_hardware_resource_open(isil_chip_t *chip)
{
    int ret = ISIL_ERR;

    atomic_inc(&chip->open_chan_number);
    if(atomic_read(&chip->open_chan_number) == 1){
        if(chip->register_timer) {
            chip->register_timer(chip, timer_on_chip_handler);
        }
    }
    ISIL_DBG(ISIL_DBG_INFO, "chip->open_chan_number = %d\n", atomic_read(&chip->open_chan_number));
    ret = ISIL_OK;
    return ret;
} 

/****************************************************************************************************************
 * Function Name :   chip_hardware_resource_close
 * Description   :   chip's hardware resource counter
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_hardware_resource_close(isil_chip_t *chip)
{
    ISIL_DBG(ISIL_DBG_INFO, "chip->open_chan_number = %d\n", atomic_read(&chip->open_chan_number));
    if(atomic_read(&chip->open_chan_number) > 0){
        if(atomic_dec_and_test(&chip->open_chan_number)){
            if(chip->unregister_timer) {
                chip->unregister_timer(chip);
            }
        }
    } else {
        ISIL_DBG(ISIL_DBG_INFO, ": open_chan_number err\n");
    }
}

/****************************************************************************************************************
 * Function Name :   chip_phy_invalid
 * Description   :   chip reset flow
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_phy_invalid(isil_chip_t *chip)
{
    u32 val;
    isil_chip_vi_driver_t *chip_vi_driver;

    ISIL_DBG(ISIL_DBG_INFO, "chip->open_chan_number = %d\n", atomic_read(&chip->open_chan_number));
#ifndef FPGA_330_5864_TESTING
    chip->io_op->chip_write32(chip, 0x8018, 0xd000 + ((chip->sys_clock << 1) - 1));
    chip->io_op->chip_write32(chip, 0x8020, 0x284);
    chip->io_op->chip_write32(chip, 0x8024, 0x20);
    chip->io_op->chip_write32(chip, 0x801c, 0x0d);
    mdelay(5);
    /*reset*/
    mpb_write(chip, ISIL_VI_SYSTEM_RESET, 0x00);/*reset*/
    mpb_write(chip, ISIL_VI_SYSTEM_RESET, 0xe0);/*unreset*/
    mpb_write(chip, ISIL_VI_SYSTEM_CLOCK, 0xf0);
    mpb_write(chip, ISIL_VI_SYSTEM_CLOCK_REVERSE, 0xf0);/*155 set to 0xff, 330 set to 0xf0*/

#if (defined(PCI_PLATFORM) && defined(DMA_INTERFACE) && defined(BURST_INTERFACE))
    chip->io_op->chip_write32(chip, MPI_CS_SETUP, 0x1);//set to 0x2 when read ping pong buffer;other set to 0x1
    chip->io_op->chip_write32(chip, MPI_RDY_HOLD, 0x3);
#else
    chip->io_op->chip_write32(chip, MPI_CS_SETUP, 0x1);//set to 0x2 when read ping pong buffer;other set to 0x1
    chip->io_op->chip_write32(chip, MPI_RDY_HOLD, 0x1);
#endif
#endif

    /*enable DDR*/
    chip->io_op->chip_write32(chip, MODE, (DAON|DDRON|CMOSON|INTRAON|INTERON|DEBLOCKINGON));

#ifndef FPGA_330_5864_TESTING
    if(chip->sys_clock > 110) {
        chip->io_op->chip_write32(chip, 0xa000, 0xc5);//Data valid counter after read command to DDR
        chip->io_op->chip_write32(chip, 0xa800, 0xc5);
    }
    /*enable i2c controler*/
    val = chip->io_op->chip_read32(chip, MCU_IIC_CONF);
    val |= 0x01;
    chip->io_op->chip_write32(chip, MCU_IIC_CONF, val);
    chip->io_op->chip_write32(chip, I2C_PHASE_SHIFT, 0x01);

    /*reset font-end*/
    chip_vi_driver = chip->chip_vi_driver;
    if(chip_vi_driver && chip_vi_driver->op) {
        chip_vi_driver->op->reset(chip_vi_driver);
    }

    /*confiure ad output 108 clock*/
    chip->io_op->chip_write32(chip, 0x8028, 0x15);
    chip->io_op->chip_write32(chip, 0x801c, 0x18);
    mdelay(5);
    chip->io_op->chip_write32(chip, 0x801c, 0x00);
#endif
}

/****************************************************************************************************************
 * Function Name :   chip_phy_valid
 * Description   :   chip init flow
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	chip_phy_valid(isil_chip_t *chip)
{
    u32 val;
    chip_driver_t  *chip_driver;

    ISIL_DBG(ISIL_DBG_INFO, "chip->open_chan_number = %d\n", atomic_read(&chip->open_chan_number));
    chip->io_op->chip_write32(chip, MODE, (DAON|DDRON|CMOSON|INTRAON|INTERON|DEBLOCKINGON));
    chip->io_op->chip_write32(chip, 0x0008, 0x0800);
    chip->io_op->chip_write32(chip, CHANNEL_SELECT, 0);
    chip->io_op->chip_write32(chip, CHANNEL_ID, 0);

    WriteForwardQuantizationTable(chip);
    WriteInverseQuantizationTable(chip);
    WriteEncodeVLCLookupTable(chip);
    ISIL_DBG(ISIL_DBG_INFO, "\nFPGA version: %x:%x, chip_id=%d\n", chip->io_op->chip_read32(chip, 0), chip->io_op->chip_read32(chip, 8), chip->chip_id);

    /*enable i2c controler*/
    val = chip->io_op->chip_read32(chip, MCU_IIC_CONF);
    val |= 0x01;
    chip->io_op->chip_write32(chip, MCU_IIC_CONF, val);
    chip->io_op->chip_write32(chip, I2C_PHASE_SHIFT, 0x01);
    chip_driver = chip->chip_driver;
    if(chip_driver && chip_driver->op){
        chip_driver->op->reset(chip_driver);
    }

#if defined(PCI_PLATFORM) && defined(DMA_INTERFACE) && defined(BURST_INTERFACE)
    /*enable burst*/
    val = chip->io_op->chip_read32(chip, PCI_MASTER_CONTROL);
    val |= (0x7 << 26);
    chip->io_op->chip_write32(chip, PCI_MASTER_CONTROL, val);
#else
    /*disable burst*/
    val = chip->io_op->chip_read32(chip, PCI_MASTER_CONTROL);
    val &= ~(0x7 << 26);
    chip->io_op->chip_write32(chip, PCI_MASTER_CONTROL, val);
#endif
}

static void chip_sysclk_init(isil_chip_t *chip, int sysclk)
{

}

/****************************************************************************************************************
 * Function Name :   chip_ddr_init
 * Description   :   chip ddr init flow
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_ddr_init(isil_chip_t *chip)
{
    chip->chip_auto_scan_ddr_size_map_mode(chip);
    chip->chip_ddr_size_map_init(chip);
}

/****************************************************************************************************************
 * Function Name :   chip_ddr_size_map_init
 * Description   :   chip ddr map param setting
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_ddr_size_map_init(isil_chip_t *chip)
{
    chip->io_op->chip_write32(chip, SET_SIZE_DDR0, chip->chip_ddr_controller1_map_mode);
    chip->io_op->chip_write32(chip, SET_SIZE_DDR1, chip->chip_ddr_controller2_map_mode);
}

/****************************************************************************************************************
 * Function Name :   chip_ddr_controller_map_mode_scan
 * Description   :   auto scan chip ddr map param
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   ddr's max page number
 ****************************************************************************************************************/
static int  chip_ddr_controller_map_mode_scan(isil_chip_t *chip, int controller)
{
    unsigned int    offset, ddrpage;

    chip->io_op->chip_write32(chip, 0x0008, 0x0000);
    chip->io_op->chip_write32(chip, MODE, DDRON);
    chip->io_op->chip_write32(chip, 0x0008, 0x0800);
    chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x6);
    chip->io_op->chip_write32(chip, DDRPAGE, (controller*0x4000));
    chip->io_op->chip_write32(chip, DDRBASE, 0x12345678);

    ddrpage = 0x400/DDR_PAGE_SIZE + (controller*0x4000);
    chip->io_op->chip_write32(chip, DDRPAGE, ddrpage);
    offset = 0x400%DDR_PAGE_SIZE;
    chip->io_op->chip_write32(chip, (DDRBASE+offset), 0x400);
    chip->io_op->chip_write32(chip, DDRPAGE, (controller*0x4000));
    if(chip->io_op->chip_read32(chip, DDRBASE) == 0x400){
        chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x4);
        ddrpage = 0x2000000/DDR_PAGE_SIZE + (controller*0x4000);
        chip->io_op->chip_write32(chip, DDRPAGE, ddrpage);
        offset = 0x2000000%DDR_PAGE_SIZE;
        chip->io_op->chip_write32(chip, (DDRBASE+offset), 0x2000000);
        if(chip->io_op->chip_read32(chip, DDRBASE) != 0x2000000){
            chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x0);
            return 64;
        } else {
            chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x4);
            return 128;
        }
    }

    ddrpage = 0x1000000/DDR_PAGE_SIZE + (controller*0x4000);
    chip->io_op->chip_write32(chip, DDRPAGE, ddrpage);
    offset = 0x1000000%DDR_PAGE_SIZE;
    chip->io_op->chip_write32(chip, (DDRBASE+offset), 0x1000000);
    chip->io_op->chip_write32(chip, DDRPAGE, (controller*0x4000));
    if(chip->io_op->chip_read32(chip, DDRBASE) == 0x1000000){
        chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x5);
        ddrpage = 0x4000000/DDR_PAGE_SIZE + (controller*0x4000);
        chip->io_op->chip_write32(chip, DDRPAGE, ddrpage);
        offset = 0x4000000%DDR_PAGE_SIZE;
        chip->io_op->chip_write32(chip, (DDRBASE+offset), 0x4000000);
        if(chip->io_op->chip_read32(chip, DDRBASE) != 0x4000000){
            chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x1);
            return 128;
        } else {
            chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x5);
            return 256;
        }
    }

    ddrpage = 0x8000000/DDR_PAGE_SIZE + (controller*0x4000);
    chip->io_op->chip_write32(chip, DDRPAGE, ddrpage);
    offset = 0x8000000%DDR_PAGE_SIZE;
    chip->io_op->chip_write32(chip, (DDRBASE+offset), 0x8000000);
    chip->io_op->chip_write32(chip, DDRPAGE, (controller*0x4000));
    if(chip->io_op->chip_read32(chip, DDRBASE) == 0x12345678){
        chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x6);
        return 512;
    }
    chip->io_op->chip_write32(chip, DDRPAGE, (controller*0x4000));
    if(chip->io_op->chip_read32(chip, DDRBASE) == 0x8000000){
        chip->io_op->chip_write32(chip, (SET_SIZE_DDR0 + controller*0x800), 0x2);
        return 256;
    }
    return 512;
}

/****************************************************************************************************************
 * Function Name :   chip_auto_scan_ddr_size_map_mode
 * Description   :   two ddr controller's mode scan
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_auto_scan_ddr_size_map_mode(isil_chip_t *chip)
{
    chip->chip_ddr_controller1_max_page_number = chip_ddr_controller_map_mode_scan(chip, 0);
    chip->chip_ddr_controller2_max_page_number = chip_ddr_controller_map_mode_scan(chip, 1);
    chip->chip_ddr_controller1_map_mode = chip->io_op->chip_read32(chip, SET_SIZE_DDR0);
    chip->chip_ddr_controller2_map_mode = chip->io_op->chip_read32(chip, SET_SIZE_DDR1);
}

/****************************************************************************************************************
 * Function Name :   chip_phy_reset
 * Description   :   chip service reset root entrance
 * Input         :   param1 is 5864's chip management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_phy_reset(isil_chip_t *chip)
{
    if(chip) {
        chip_driver_t       *chip_driver = chip->chip_driver;
        isil_chip_ai_driver_t *chip_ai_driver = chip->chip_ai_driver;
        isil_chip_vi_driver_t *chip_vi_driver = chip->chip_vi_driver;

        if(chip_ai_driver && chip_ai_driver->op) {
            chip_ai_driver->op->reset(chip_ai_driver);
        }
        if(chip_vi_driver && chip_vi_driver->op) {
            chip_vi_driver->op->reset(chip_vi_driver);
        }
        if(chip_driver && chip_driver->op) {
            chip_driver->op->reset(chip_driver);
        }
    }
}

/****************************************************************************************************************
 * Function Name :   register_curr_h264_encode_chan
 * Description   :   register 264 encode's isr service
 * Input         :   param1 is 5864's chip management struct.
 *                   param2 is 264 logic chan's management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void register_curr_h264_encode_chan(isil_chip_t *chip, isil_h264_logic_encode_chan_t *h264_logic_encode_chan)
{
    if((chip!=NULL) && (h264_logic_encode_chan!=NULL)){
        chip->vlc_ping_pong_index = 0;
        chip_request_irq(chip, IRQ_VLC_TYPE_INTR, h264_logic_encode_chan->encode_control.op->irq_func, "VLC", (void*)&h264_logic_encode_chan->encode_control);
    }
}

/****************************************************************************************************************
 * Function Name :   unregister_curr_h264_encode_chan
 * Description   :   unregister 264 encode's isr service
 * Input         :   param1 is 5864's chip management struct.
 *                   param2 is 264 logic chan's management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void	unregister_curr_h264_encode_chan(isil_chip_t *chip, isil_h264_logic_encode_chan_t *h264_logic_encode_chan)
{
    if((chip!=NULL) && (h264_logic_encode_chan!=NULL)){
        chip_free_irq(chip, IRQ_VLC_TYPE_INTR, &h264_logic_encode_chan->encode_control);
    }
}

/****************************************************************************************************************
 * Function Name :   register_curr_h264_encode_chan
 * Description   :   register 264 encode's isr service
 * Input         :   param1 is 5864's chip management struct.
 *                   param2 is 264 logic chan's management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void register_curr_h264_encode_push_chan(isil_chip_t *chip, isil_h264_logic_encode_chan_t *h264_logic_encode_chan)
{
    if((chip!=NULL) && (h264_logic_encode_chan!=NULL)){
        chip_request_irq(chip, IRQ_VLC_DONE_TYPE_INTR, h264_logic_encode_chan->encode_control.op->irq_func, "VLC PUSH", (void*)&h264_logic_encode_chan->encode_control);
    }
}

/****************************************************************************************************************
 * Function Name :   unregister_curr_h264_encode_chan
 * Description   :   unregister 264 encode's isr service
 * Input         :   param1 is 5864's chip management struct.
 *                   param2 is 264 logic chan's management struct
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void unregister_curr_h264_encode_push_chan(isil_chip_t *chip, isil_h264_logic_encode_chan_t *h264_logic_encode_chan)
{
    if((chip!=NULL) && (h264_logic_encode_chan!=NULL)){
        chip_free_irq(chip, IRQ_VLC_DONE_TYPE_INTR, &h264_logic_encode_chan->encode_control);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_init
 * Description   :   chip 264 encode schedule queue service
 * Input         :   param1 is chip encode schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_encode_chan_service_queue_init(chip_encode_chan_service_queue_t *chip_encode_service_queue)
{
    if(chip_encode_service_queue != NULL){
        tcb_node_queue_t    *queue_node = &chip_encode_service_queue->service_queue_node;
        spin_lock_init(&chip_encode_service_queue->chip_encode_chan_service_queue_lock);
        atomic_set(&chip_encode_service_queue->h264_master_can_recv_number, 2);
        atomic_set(&chip_encode_service_queue->h264_ddr_cache_number, 8);
        queue_node->op = &tcb_node_queue_op;
        queue_node->op->init(queue_node);
        chip_encode_service_queue->curr_consumer = NULL;
    }
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_get_queue_curr_entry_number
 * Description   :   get schedule queue's prepare service entry's number
 * Input         :   param1 is chip encode schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int  chip_encode_chan_service_queue_get_queue_curr_entry_number(chip_encode_chan_service_queue_t *chip_encode_service_queue)
{
    int ret = 0;
    if(chip_encode_service_queue != NULL){
        tcb_node_queue_t    *queue_node = &chip_encode_service_queue->service_queue_node;
        ret = queue_node->op->get_queue_curr_entry_number(queue_node);
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_put_service_request_into_queue
 * Description   :   put one entry into schedule queue from tail
 * Input         :   param1 is chip encode schedule queue management struct.
 *                   param2 entry of schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_encode_chan_service_queue_put_service_request_into_queue(chip_encode_chan_service_queue_t *chip_encode_service_queue, encode_chan_service_tcb_t *chan_service_tcb)
{
    if((chip_encode_service_queue!=NULL) && (chan_service_tcb!=NULL)){
        tcb_node_queue_t    *queue_node = &chip_encode_service_queue->service_queue_node;
        queue_node->op->put(queue_node, &chan_service_tcb->service_tcb);
        chip_encode_service_queue->op->trigger_chip_pending_service_request(chip_encode_service_queue);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_put_service_request_into_queue_header
 * Description   :   put one entry into schedule queue from header
 * Input         :   param1 is chip encode schedule queue management struct.
 *                   param2 entry of schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_encode_chan_service_queue_put_service_request_into_queue_header(chip_encode_chan_service_queue_t *chip_encode_service_queue, encode_chan_service_tcb_t *chan_service_tcb)
{
    if((chip_encode_service_queue!=NULL) && (chan_service_tcb!=NULL)){
        tcb_node_queue_t    *queue_node = &chip_encode_service_queue->service_queue_node;
        queue_node->op->put_header(queue_node, &chan_service_tcb->service_tcb);
        chip_encode_service_queue->op->trigger_chip_pending_service_request(chip_encode_service_queue);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_delete_service_request_from_queue
 * Description   :   delete one entry from schedule queue
 * Input         :   param1 is chip encode schedule queue management struct.
 *                   param2 entry of schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static int  chip_encode_chan_service_queue_delete_service_request_from_queue(chip_encode_chan_service_queue_t *chip_encode_service_queue, encode_chan_service_tcb_t *chan_service_tcb)
{
    int ret = 0;
    if((chip_encode_service_queue!=NULL) && (chan_service_tcb!=NULL)){
        tcb_node_queue_t    *queue_node = &chip_encode_service_queue->service_queue_node;
        unsigned long       flags;
        spin_lock_irqsave(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);
        ret = queue_node->op->delete(queue_node, &chan_service_tcb->service_tcb);
        if(ret == 1){
            if(chan_service_tcb->delete_req_notify != NULL){
                chan_service_tcb->delete_req_notify(chan_service_tcb->context);
            }
        } 
        spin_unlock_irqrestore(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);
    }
    return ret;
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_try_get_curr_consumer_from_queue
 * Description   :   try get one entry from schedule queue
 * Input         :   param1 is chip encode schedule queue management struct.
 *                   param2 entry of schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_encode_chan_service_queue_try_get_curr_consumer_from_queue(chip_encode_chan_service_queue_t *chip_encode_service_queue)
{
    if(chip_encode_service_queue != NULL){
        tcb_node_queue_t    *queue_node = &chip_encode_service_queue->service_queue_node;

        if(chip_encode_service_queue->curr_consumer == NULL){
            if(queue_node->op != NULL){
                tcb_node_t  *temp_node;
                queue_node->op->try_get(queue_node, &temp_node);
                if(temp_node != NULL){
                    chip_encode_service_queue->curr_consumer = to_get_encode_chan_service_tcb_with_service_tcb(temp_node);
                }
            }
        }
    }
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_release_curr_consumer
 * Description   :   release curr consumer entry of schedule queue
 * Input         :   param1 is chip encode schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_encode_chan_service_queue_release_curr_consumer(chip_encode_chan_service_queue_t *chip_encode_service_queue)
{
    if(chip_encode_service_queue != NULL){
        unsigned long	flags;
        spin_lock_irqsave(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);
        chip_encode_service_queue->curr_consumer = NULL;
        spin_unlock_irqrestore(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);
    }
}

/****************************************************************************************************************
 * Function Name :   chip_encode_chan_service_queue_trigger_chip_pending_service_request
 * Description   :   trigger schedule queue to run
 * Input         :   param1 is chip encode schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void chip_encode_chan_service_queue_trigger_chip_pending_service_request(chip_encode_chan_service_queue_t *chip_encode_service_queue)
{
    if(chip_encode_service_queue != NULL){
        isil_chip_t *chip = to_get_chip_with_chip_encode_service_queue(chip_encode_service_queue);
        encode_chan_service_tcb_t   *curr_consumer;
        unsigned long	flags;
        int curr_entry_number;

again:
        spin_lock_irqsave(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);
        curr_entry_number = chip_encode_service_queue->op->get_queue_curr_entry_number(chip_encode_service_queue);
        if((curr_entry_number==0) && (chip_encode_service_queue->curr_consumer==NULL)){
            spin_unlock_irqrestore(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);
            return;
        } else if(chip_encode_service_queue->curr_consumer != NULL){
            spin_unlock_irqrestore(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);
            return;
        }
        chip_encode_service_queue->op->try_get_curr_consumer_from_queue(chip_encode_service_queue);
        curr_consumer = chip_encode_service_queue->curr_consumer;
        spin_unlock_irqrestore(&chip_encode_service_queue->chip_encode_chan_service_queue_lock, flags);

        if(curr_consumer != NULL){
            if(chip_is_in_robust_processing(chip)){
                if(curr_consumer->delete_req_notify != NULL){
                    curr_consumer->delete_req_notify(curr_consumer->context);
                }
                chip_encode_service_queue->op->release_curr_consumer(chip_encode_service_queue);
                goto again;
            } else {
                if(curr_consumer->req_callback != NULL){
                    curr_consumer->req_callback(curr_consumer->context);                
                } else {
                    if(curr_consumer->delete_req_notify != NULL){
                        curr_consumer->delete_req_notify(curr_consumer->context);
                    }
                    chip_encode_service_queue->op->release_curr_consumer(chip_encode_service_queue);
                    ISIL_DBG(ISIL_DBG_INFO, ": &&&&&&&&&&&&&,%d\n\n", curr_entry_number);
                    goto again;
                }
            }
        }
    }
}

struct chip_encode_chan_service_queue_operation chip_encode_chan_service_queue_op = {
    .init = chip_encode_chan_service_queue_init,
    .get_queue_curr_entry_number = chip_encode_chan_service_queue_get_queue_curr_entry_number,
    .put_service_request_into_queue = chip_encode_chan_service_queue_put_service_request_into_queue,
    .put_service_request_into_queue_header = chip_encode_chan_service_queue_put_service_request_into_queue_header,
    .delete_service_request_from_queue = chip_encode_chan_service_queue_delete_service_request_from_queue,
    .try_get_curr_consumer_from_queue = chip_encode_chan_service_queue_try_get_curr_consumer_from_queue,
    .release_curr_consumer = chip_encode_chan_service_queue_release_curr_consumer,
    .trigger_chip_pending_service_request = chip_encode_chan_service_queue_trigger_chip_pending_service_request,
};

/****************************************************************************************************************
 * Function Name :   init_chip_encode_chan_service_queue
 * Description   :   init 264 encode's schedule queue
 * Input         :   param1 is chip encode schedule queue management struct.
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static void init_chip_encode_chan_service_queue(chip_encode_chan_service_queue_t * chip_encode_service_queue)
{
    if(chip_encode_service_queue != NULL){
        chip_encode_service_queue->op = &chip_encode_chan_service_queue_op;
        chip_encode_service_queue->op->init(chip_encode_service_queue);
    }
}

static u32  isil_host_io_chip_read32(isil_chip_t *chip, u32 offset)
{
    return __raw_readl((chip->regs+offset));
}

static u16  isil_host_io_chip_read16(isil_chip_t *chip, u32 offset)
{
    return __raw_readw((chip->regs+offset));
}

static u8  isil_host_io_chip_read8(isil_chip_t *chip, u32 offset)
{
    return __raw_readb((chip->regs+offset));
}

static void  isil_host_io_chip_read_block(isil_chip_t *chip, u32 offset, int len, u32 *dest)
{
    int i;
    len += 3;
    len >>= 2;
    for(i=0; i<len; i++){
        dest[i] = __raw_readl((chip->regs+offset));
        offset += 4;
    }
}

static void isil_host_io_chip_write32(isil_chip_t *chip, u32 offset, u32 value)
{
    __raw_writel(value, (chip->regs+offset));
#if 0
    if(offset < 0x2000){
        printk("264: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0x3000){
        //printk("bitstream: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0x4000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0x5000){
        printk("audio: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0x802c){
        printk("system_config: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0x8800){
        printk("reserver space %x\n", offset);
    } else if(offset < 0x8900){
        printk("Int: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0x9000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0x9210){
        printk("VIF: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0xa000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0xa900){
        printk("DDR: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0xb800){
        printk("reserver space %x\n", offset);
    } else if(offset < 0xb810){
        printk("IAR: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0xc000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0xc800){
        printk("preview: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0xc808){
        printk("MJPEG_Capture: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0xd000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0xd100){
        printk("MJPEG_Control: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0xe000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0xfc00){
        printk("Motion_vector: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0x18000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0x18200){
        printk("PCI_Control_map: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else if(offset < 0x80000){
        printk("reserver space %x\n", offset);
    } else if(offset < 0x88000){
        printk("ddr_burst_map: %x=0x%x, 0x%x\n", offset, value, __raw_readl(chip->regs+offset));
    } else {
        printk("reserver space %x\n", offset);    
    }
#endif
}

static void isil_host_io_chip_write16(isil_chip_t *chip, u32 offset, u16 value)
{
    __raw_writew(value, (chip->regs+offset));
}

static void isil_host_io_chip_write8(isil_chip_t *chip, u32 offset, u8 value)
{
    __raw_writeb(value, (chip->regs+offset));
}

static void isil_host_io_chip_write_block(isil_chip_t *chip, u32 offset, int len, u32 *src)
{
    int i;
    len += 3;
    len >>= 2;
    for(i=0; i<len; i++){
        __raw_writel(src[i], (chip->regs+offset));
        offset += 4;
    }
}

static __used struct chip_io_operation isil_host_io = {
    .chip_read32 = isil_host_io_chip_read32,
    .chip_read16 = isil_host_io_chip_read16,
    .chip_read8 = isil_host_io_chip_read8,
    .chip_read_block = isil_host_io_chip_read_block,
    .chip_write32 = isil_host_io_chip_write32,
    .chip_write16 = isil_host_io_chip_write16,
    .chip_write8 = isil_host_io_chip_write8,
    .chip_write_block = isil_host_io_chip_write_block,
};

#ifdef POWERPC_PLATFORM
static u32  isil_pci_io_chip_read32(isil_chip_t *chip, u32 offset)
{
    return in_le32((chip->regs+offset));
}

static u16  isil_pci_io_chip_read16(isil_chip_t *chip, u32 offset)
{
    return in_le16((chip->regs+offset));
}

static u8  isil_pci_io_chip_read8(isil_chip_t *chip, u32 offset)
{
    return in_8((chip->regs+offset));
}

static void  isil_pci_io_chip_read_block(isil_chip_t *chip, u32 offset, int len, u32 *dest)
{
    int i;

    len += 3;
    len >>= 2;
    for(i=0; i<len; i++){
        dest[i] = in_le32((chip->regs+offset));
        offset += 4;
    }
}

static void isil_pci_io_chip_write32(isil_chip_t *chip, u32 offset, u32 value)
{
    out_le32((chip->regs+offset), value);
}

static void isil_pci_io_chip_write16(isil_chip_t *chip, u32 offset, u16 value)
{
    out_le16((chip->regs+offset), value);
}

static void isil_pci_io_chip_write8(isil_chip_t *chip, u32 offset, u8 value)
{
    out_8((chip->regs+offset), value);
}

static void isil_pci_io_chip_write_block(isil_chip_t *chip, u32 offset, int len, u32 *src)
{
    int i;

    len += 3;
    len >>= 2;
    for(i=0; i<len; i++){
        out_le32((chip->regs+offset), src[i]);
        offset += 4;
    }
}

static struct chip_io_operation isil_pci_io = {
    .chip_read32 = isil_pci_io_chip_read32,
    .chip_read16 = isil_pci_io_chip_read16,
    .chip_read8 = isil_pci_io_chip_read8,
    .chip_read_block = isil_pci_io_chip_read_block,
    .chip_write32 = isil_pci_io_chip_write32,
    .chip_write16 = isil_pci_io_chip_write16,
    .chip_write8 = isil_pci_io_chip_write8,
    .chip_write_block = isil_pci_io_chip_write_block,
};
#endif

/****************************************************************************************************************
 * Function Name :   chip_proc_read
 * Description   :   help function for debugging
 * Input         :   param1 is seq_file ptr
 *                   param2 is register context
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static ssize_t chip_proc_read(struct seq_file *seq, void *data)
{
    u32 i;
    DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY *video_encode_cap;
    isil_chip_t *chip = (isil_chip_t *)seq->private;
    isil_chip_timer_controller_t    *chip_timer_cont;
    TMCB        *ptr_singletimer;
    FORTMCB     *ptr_fortimer;

    video_encode_cap = chip->video_encode_cap;
    seq_printf(seq, "Version: %d.%d.%d-%d\nGCC Version: %s\nARCH: %s\nBuild at: __DATE__ __TIME__ \n", 
            get_version_major(__ISIL__CODE__VERSION__),
            get_version_submajor(__ISIL__CODE__VERSION__),
            get_version_minor(__ISIL__CODE__VERSION__), 
            __ISIL_SVN_VERSION__,__VERSION__, __ISIL_ARCH__ );
    seq_printf(seq, "device name \"%s\", bus %d, chip_id %d, sys_clk = %dMhz, map_mode = %d\n", chip->name, chip->bus_id, chip->chip_id, chip->sys_clock, video_encode_cap->ddr_map_mode);
    seq_printf(seq, "base: 0x%08x, size: 0x%08x\n", video_encode_cap->fpga_async_base_addr, video_encode_cap->fpga_async_section_len);

    seq_printf(seq, "\n");
    seq_printf(seq, "%8s %8s %8s %8s %8s %8s %8s %8s %8s %8s %8s\n",
            "irq", "name", "n_irq", "lost", "depth", "irq_base", "irq_delta", "duration(us)", "total_duration(us)", "system_duratoion(ms)", "irq_ratio");
    chip->isil_int_control.system_op->calc_irq_ratio_with_system(&chip->isil_int_control, chip);
    for(i = 0; i < CHIP_IRQS_NR; i++)
    {
        struct chip_irq_desc *desc = chip_irq_to_desc(chip, i);
        if(desc/* && desc->action.handler*/){
            seq_printf(seq, "%8d %8s %8u %8u %8d %8u %8u %8u %8llu %8llu %8u\n",
                    desc->irq, desc->name, desc->irq_count,
                    desc->irqs_unhandled, atomic_read(&desc->depth),
                    desc->base_irq_count, desc->delta_irq_count,
                    desc->irq_duration,
                    desc->irq_total_duration,
                    chip->isil_int_control.system_op->get_system_duration(&chip->isil_int_control),
                    chip->isil_int_control.system_op->get_irq_ratio(&chip->isil_int_control));
        }
    }

    chip_timer_cont = &chip->chip_timer_cont;
    ptr_singletimer = chip_timer_cont->SingleFireQueue;
    ptr_fortimer = chip_timer_cont->ForFireQueue;
    seq_printf(seq, "\n");
    seq_printf(seq, "single timer trigger\n");
    seq_printf(seq, "%8s %8s %8s\n", "name", "duration(us)", "total(us)");
    for(i=0; i<MAXSINGLEFIREQUEUENUM; i++){
        if(ptr_singletimer->flag == TIMER_VALIDE){
            seq_printf(seq, "%pS %8u %8llu\n", ptr_singletimer->hook, ptr_singletimer->timer_duration, ptr_singletimer->timer_total_duration);
        }
        ptr_singletimer++;
    }
    seq_printf(seq, "\n");
    seq_printf(seq, "for timer trigger\n");
    seq_printf(seq, "%8s %8s %8s\n", "name", "duration(us)", "total(us)");
    for(i=0; i<MAXSINGLEFIREQUEUENUM; i++){
        if(ptr_fortimer->flag == TIMER_VALIDE){
            seq_printf(seq, "%pS %8u %8llu\n", ptr_fortimer->hook, ptr_fortimer->timer_duration, ptr_fortimer->timer_total_duration);
        }
        ptr_fortimer++;
    }
    return 0;
}

static int inline gen_verify_data(int test_case, u32 data)
{
    switch(test_case) {
        default:
        case 0:return data;
        case 1:return ~data;
        case 2:return ((data & 0x1) ? 0xaa55aa55 : 0x55aa55aa);
    }
}

#if defined(PCI_PLATFORM)
/****************************************************************************************************************
 * Function Name :   chip_proc_write
 * Description   :   help function for debugging
 * Input         :   
 * Output        :   null
 * Return        :   null
 ****************************************************************************************************************/
static ssize_t chip_proc_write(struct file *file, const char __user *buffer, size_t count, loff_t *offset)
{
    isil_chip_t *chip = (isil_chip_t *)isil_file_get_seq_private(file);
    char cmdbuf[128];
    char **argv;
    int argc, i,cnt;
    u32 val = 0;

    if(!chip){
        ISIL_DBG(ISIL_DBG_INFO, "internal error!\n");
        return 0;
    }
    if(count > 128){
        if(copy_from_user(cmdbuf, buffer, 128) != 0){
            ISIL_DBG(ISIL_DBG_INFO, ": copy_from_user failed\n");
        }
    }
    else{
        if(copy_from_user(cmdbuf, buffer, count) != 0){
            ISIL_DBG(ISIL_DBG_INFO, ": copy_from_user failed\n");
        }
    }
    argv = argv_split(GFP_KERNEL, cmdbuf, &argc);
    for(i = 0; i < (argc - 1); i++)
    {
        ISIL_DBG(ISIL_DBG_INFO, "cmd %d: %s\n", i, argv[i]);
    }
    if((argv[0] != NULL) && (strcmp(argv[0], "burst") == 0))
    {
        chip->io_op->chip_write32(chip, 0x18004, 0x7c000000);
        val = chip->io_op->chip_read32(chip, ENABLE_VLC_LOOKUP_TABLE) | 0x2;
        chip->io_op->chip_write32(chip, ENABLE_VLC_LOOKUP_TABLE, val);
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "single") == 0))
    {
        chip->io_op->chip_write32(chip, 0x18004, 0x7c000000);
        chip->io_op->chip_write32(chip, ENABLE_VLC_LOOKUP_TABLE, chip->io_op->chip_read32(chip, ENABLE_VLC_LOOKUP_TABLE) | 0x2);
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "h264_burst") == 0))
    {
        //chip->io_op->chip_write32(chip,MODE, DDRON);//enable ddr
        while(1)isil_vlc_irq_handler(0, chip);
    }
    /*close ad*/
    else if((argv[0] != NULL) && (strcmp(argv[0], "poweroff") == 0)){
        pci_i2c_write(chip, 0x50, 0xce, 0x3f);
        pci_i2c_write(chip, 0x52, 0xce, 0x3f);
        pci_i2c_write(chip, 0x54, 0xce, 0x3f);
        pci_i2c_write(chip, 0x56, 0xce, 0x3f);
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "poweron") == 0)){
        pci_i2c_write(chip, 0x50, 0xce, 0x00);
        pci_i2c_write(chip, 0x52, 0xce, 0x00);
        pci_i2c_write(chip, 0x54, 0xce, 0x00);
        pci_i2c_write(chip, 0x56, 0xce, 0x00);
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "mpb_read") == 0)){
        printk("\nread 0x%08x == >0x%08x\n", atoi(argv[1]), mpb_read(chip, atoi(argv[1])));
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "mpb_write") == 0)){
        mpb_write(chip, atoi(argv[1]), atoi(argv[2]));
        printk("\nwrite 0x%08x == >0x%08x\n", atoi(argv[1]), atoi(argv[2]));
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "mpb_dump") == 0)){
        u32 base;
        u32 length;

        if(argc >= 3){
            base = atoi(argv[1]);
            length = atoi(argv[2]);
            for(i = 0; i < length; i++){
                if((i % 16) == 0){
                    printk("\n0x%08x:", base + i);
                }
                printk(" %02x", mpb_read(chip, base + i));
            }
            printk("\n");
        }
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "i2c_read") == 0)){
        u8 temp = 0;
        pci_i2c_read(chip, atoi(argv[1]), atoi(argv[2]), &temp);
        printk("\nread 0x%08x, 0x%08x == >0x%x\n", atoi(argv[1]), atoi(argv[2]), temp);
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "i2c_write") == 0)){
        pci_i2c_write(chip, atoi(argv[1]), atoi(argv[2]), atoi(argv[3]));
        printk("\nwrite 0x%08x, 0x%08x == >0x%08x\n", atoi(argv[1]), atoi(argv[2]), atoi(argv[3]));
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "i2c_dump") == 0)){
        int iic = atoi(argv[1]);
        u8 val = 0;

        printk("0x%02x", iic);
        for(i = 0; i < 256; i++) {
            if((i % 16) == 0) {
                printk("\n%08x:", i);
            }
            pci_i2c_read(chip, iic, i, &val);
            printk("%02x ", val);
        }
        printk("\n");
    }else if((argv[0] != NULL) && (strcmp(argv[0], "dump") == 0) && (argc > 4)){
        struct file *mem_file;
        mm_segment_t fs; 
        int ab_switch = 0;
        u32 ddr_start, size, page_id, offset, page_cnt, len;
        char *buff, *temp_buf, *p_buff;
        if((argv[1][0] == 'A') || (argv[1][0] == 'a')){
            ab_switch = 0;
        }else{
            ab_switch = 1;
        }
        temp_buf = (char*)__get_free_pages(GFP_KERNEL, get_order(0x80000));
        p_buff = temp_buf;
        ddr_start = atoi(argv[2]);
        size      = atoi(argv[3]);
        page_id = (ddr_start>>19) & 0xff;
        offset  = ddr_start & 0x0007ffff;
        page_cnt = (size / 0x80000);
        printk("save ddr %s: begin 0x%08x size 0x%08x , start page id %d offset 0x%08x...\n", ab_switch?"B":"A", ddr_start, size, page_id, offset);
        mem_file = isil_kernel_file_open(argv[4]);
        fs=get_fs();
        set_fs(KERNEL_DS);
        chip->io_op->chip_write32(chip, 0x04, 0x1);
        for(i = 0; i < page_cnt; i++){
            /*switch page*/
            chip->io_op->chip_write32(chip, 0x30, (ab_switch << 14) | (page_id + i));
            if(i == 0){
                buff = chip->regs + 0x80000 + offset;
                len = 0x80000 - offset;
            }else{
                buff = chip->regs + 0x80000;
                len = 0x80000;
            }
            printk("--->write size %08x\n", len);
            /*write to file*/
            mem_file->f_op->write(mem_file, buff, len, &(mem_file->f_pos));
        }
        len = (size % 0x80000);
        if(len){
            chip->io_op->chip_write32(chip, 0x30, (ab_switch << 14) | (page_id + i));
            if(page_cnt == 0){
                buff = chip->regs + 0x80000 + offset;
            }else{
                buff = chip->regs + 0x80000;
            }
            printk("--->write size %08x\n", len);
            /*write tailer*/
            mem_file->f_op->write(mem_file, buff, len, &(mem_file->f_pos));
        }
        set_fs(fs);
        isil_kernel_file_close(mem_file);
        free_pages((unsigned long)temp_buf, get_order(0x80000));
    }
    else if((argv[0] != NULL) && (strcmp(argv[0], "ddr_test") == 0))
    {
        u32 *dma_vir = NULL;
        u32 *dma_vir_temp = NULL;
        int error = 0;
        int test_flags = 0;
        int retry_cnt = 0, data_base = 0;
        int page_id = 0, ab_switch = 0, ddr_a_size = chip->chip_ddr_controller1_max_page_number<<19, ddr_b_size = chip->chip_ddr_controller1_max_page_number<<19;
        int test_case = 0;

        u32 section_size = 0x800;
        u32 section_cnt;
        chip_ddr_burst_engine_t *chip_ddr_burst_interface;
        ddr_burst_interface_t   *burst_interface;
        dpram_page_node_t       *dpram_page;

        chip_ddr_burst_interface = &chip->chip_ddr_burst_interface;
        chip_ddr_burst_interface->op->get_ddr_burst_interface(chip_ddr_burst_interface, &burst_interface);
#define DDR_PIO    0x0001
#define DDR_BURST  0x0002
#define SRAM_PIO   0x0004
#define SRAM_BURST 0x0008


        for(i = 2; i < argc; i++){
            if(strcmp(argv[i], "ddr_pio") == 0){
                test_flags |= DDR_PIO;
            }
            if(strcmp(argv[i], "ddr_burst") == 0){
                test_flags |= DDR_BURST;
            }
            if(strcmp(argv[i], "sram_pio") == 0){
                test_flags |= SRAM_PIO;
            }
            if(strcmp(argv[i], "sram_burst") == 0){
                test_flags |= SRAM_BURST;
            }
            if(strcmp(argv[i], "ddr_a_size") == 0){
                ddr_a_size = atoi(argv[i + 1]);
            }
            if(strcmp(argv[i], "ddr_b_size") == 0){
                ddr_b_size = atoi(argv[i + 1]);
            }
        }
        retry_cnt = atoi(argv[1]);
        printk("--->0x%08x<---retry %d\n", test_flags, retry_cnt);
        dma_vir = (u32 *)__get_free_pages(GFP_KERNEL, 0);
        if(!dma_vir)
        {
            printk("no memory!\n");
            goto ddr_test_end;
        }
        memset(dma_vir, 0, PAGE_SIZE);
        dma_vir_temp = (u32 *)__get_free_pages(GFP_KERNEL, 0);
        if(!dma_vir)
        {
            printk("no memory!\n");
            goto ddr_test_end;
        }
        memset(dma_vir_temp, 0, PAGE_SIZE);
        /*test all pages*/
        //chip->io_op->chip_write32(chip, MODE, DDRON);//enable ddr
        //chip->io_op->chip_write32(chip, PCI_MASTER_CONTROL, 0x1c000000);
        chip->io_op->chip_write32(chip, MPI_CS_SETUP, 0x1);//set to 0x2 when read ping pong buffer;other set to 0x1
        chip->io_op->chip_write32(chip, MPI_RDY_HOLD, 0x3);
        data_base = 0;
        page_id = 0;
        ab_switch = 0;
        test_case = 0;
        while((retry_cnt) >= 0) {
            if(page_id >= 512) {
                page_id = 0;
                retry_cnt--;
                if(retry_cnt <= 0){
                    break;
                }
            }
            data_base = (1 << 19) * page_id;
            if(test_flags & DDR_BURST) {
                u32 ddr_end_addr;
                u32 page_size;
                u32 *temp_pio_addr;
                u32 *gen_data;

                page_size = 0x80000;
                section_size = 0x800;
                cnt = 0;
                section_cnt = page_size/section_size;
                while((cnt < section_cnt) && chip_ddr_burst_interface->op->is_can_submit_move_data_from_host_to_dpram_service_req(chip_ddr_burst_interface, &dpram_page)){
                    ddr_end_addr = data_base + section_size * cnt;

                    gen_data = dma_vir;
                    while((gen_data - dma_vir) < (section_size >> 2)) {
                        *gen_data = gen_verify_data(test_case, ddr_end_addr + (gen_data - dma_vir));
                        gen_data++;
                    }
                    temp_pio_addr = (u32 *)dma_vir;

                    burst_interface->op->pio_host_to_sram_write(burst_interface, dpram_page, (u32*)temp_pio_addr, section_size, 0);
                    burst_interface->op->start_block_transfer_from_sram_to_ddr(burst_interface, dpram_page, 
                            ddr_end_addr, page_id, section_size, DDR_CHIP_A);
                    chip_ddr_burst_interface->op->notify_end_move_data_from_dpram_to_ddr_service_req(chip_ddr_burst_interface, &dpram_page);

                    cnt++;
                }

                cnt = 0;
                while((cnt < section_cnt) && chip_ddr_burst_interface->op->is_can_submit_move_data_from_ddr_to_dpram_service_req(chip_ddr_burst_interface, &dpram_page)){
                    ddr_end_addr = data_base + section_size * cnt;
                    temp_pio_addr = (u32 *) dma_vir_temp;

                    burst_interface->op->start_block_transfer_from_ddr_to_sram(burst_interface, dpram_page, 
                            ddr_end_addr, page_id, section_size, DDR_CHIP_A);
                    burst_interface->op->pio_host_to_sram_read(burst_interface, dpram_page, (u32*)temp_pio_addr, section_size, 0);
                    chip_ddr_burst_interface->op->notify_end_move_data_from_dpram_to_host_service_req(chip_ddr_burst_interface, &dpram_page);
                    cnt++;
                    gen_data = dma_vir_temp;
                    while((gen_data - dma_vir_temp) < (section_size >> 2)) {
                        if(*gen_data != gen_verify_data(test_case, ddr_end_addr + (gen_data - dma_vir_temp))){
                            error = 1;
                            printk("verfiy data failed! expect=0x%08x, readback = 0x%08x\n", gen_verify_data(test_case, ddr_end_addr + (gen_data - dma_vir_temp)), *gen_data);
                            schedule();
                            if (signal_pending(current)) {
                                goto ddr_test_end;
                            }
                        }
                        gen_data++;
                    }
                }

            }
            test_case++;
            if(test_case > 2) {
                page_id++;

                test_case = 0;
                printk("%s page %3d %s\n", ab_switch?"DDRB":"DDRA", page_id, error?"FAILED":"OK");
                error = 0;

                if(page_id == (ddr_a_size >> 19)){
                    page_id = 256;
                }else if(page_id == ((ddr_b_size >> 19) + 256)){
                    page_id = 512;
                }
            }
            schedule();
            if (signal_pending(current)) {
                goto ddr_test_end;
            }
        }
ddr_test_end:
        if(dma_vir) {
            free_page((unsigned long)dma_vir);
        }
        if(dma_vir_temp) {
            free_page((unsigned long)dma_vir_temp);
        }
    }
    argv_free(argv);

    return count;
}
#endif

//struct isil_chip_device * tcd = NULL;

/************************************************************************************/
/************************************************************************************/


int register_all_tc_device(void)
{
    /*all the functions blew have handle error, so the return value can be ignore*/
    isil_chip_ai_init();
    isil_chip_vi_init();
    isil_audio_en_init();
    isil_audio_de_init();
#if defined(PCI_PLATFORM)
    isil_jpeg_en_init();
#endif
    isil_videoh264_en_init();
    isil_videoh264s_en_init();
    return ISIL_OK;
}

void unregister_all_tc_devices(void)
{
    isil_chip_vi_remove();
    isil_chip_ai_remove();
    isil_audio_en_remove();
    isil_audio_de_remove();
#if defined(PCI_PLATFORM)
    isil_jpeg_en_remove();
#endif
    isil_videoh264_en_remove();
    isil_videoh264s_en_remove();
}


int register_all_tc_driver(void)
{
    int ret = -1;

    if(isil_chip_ai_driver_init()){
        return ret;
    }
    if(isil_chip_vi_driver_init()){
        goto err1;
    }
    if(isil_audio_en_driver_init()) {  
        goto err2;
    }
    if(isil_audio_de_driver_init()){
        goto err3;
    }
    if(isil_video_en_driver_init()){
        goto err4;
    }
    if(isil_videosub_en_driver_init()){
        goto err5;
    }
#if defined(PCI_PLATFORM)
    if(isil_jpeg_en_driver_init()){
        goto err6;
    }
#endif

    return 0;

#if defined(PCI_PLATFORM)
err6:
    isil_videosub_en_driver_remove();
#endif
err5:
    isil_video_en_driver_remove();
err4:
    isil_audio_de_driver_remove();
err3:
    isil_audio_en_driver_remove();
err2:
    isil_chip_vi_driver_remove();
err1:
    isil_chip_ai_driver_remove();

    return -1;
}

void unregister_all_tc_driver(void)
{
    isil_chip_ai_driver_remove();
    isil_chip_vi_driver_remove();
    isil_audio_de_driver_remove();
    isil_audio_en_driver_remove();
    isil_video_en_driver_remove();
    isil_videosub_en_driver_remove();
#if defined(PCI_PLATFORM)
    isil_jpeg_en_driver_remove();
#endif
}




static void isil_chip_remove(struct isil_chip_device * tcd)
{
    isil_chip_t *chip = tcd_priv(tcd);

    if(chip == NULL){
        return;
    }
    remove_isil_interrupt_control(&chip->isil_int_control, chip);
    if(chip->regs != NULL){
        iounmap(chip->regs);
        chip->regs=NULL;
    }
    if(chip->resource != NULL){
        release_resource(chip->resource);
        kfree(chip->resource);
        chip->resource = NULL;
    }
    if(chip->sync_regs != NULL){
        iounmap(chip->sync_regs);
        chip->sync_regs = NULL;
    }
    if(chip->sync_resource != NULL){
        release_resource(chip->sync_resource);
        kfree(chip->sync_resource);
        chip->sync_resource = NULL;
    }

    //remove chip_bus
    remove_chip_bus(&chip->isil_chip_bus, chip->bus_id);
    isil_module_unregister(chip, &chip->chip_proc);
#if defined(PCI_PLATFORM)
    push_buf_pool_release(chip);
#endif
}

void remove_chip_device(struct isil_chip_device * tcd)
{
    //printk("before isil_chip_exit tcd[0x%p]\n",tcd);
    tc_chip_exit(tcd); 
}

void release_chip_memory(struct isil_chip_device * tcd)
{
    tc_chipdev_free(tcd);
    //printk("after tc_chip_dev_free\n");

}

/****************************************************************************************************************
 * Function Name :   isil_chip_init
 * Description   :   this function is tw5864 chip's root init function.
 * Input         :   param1 description the chip have been connected pci bus or host bus, 
 *                   param2 is 5864's logic id,
 *                   param3 is capbility of this chip,
 *                   param4 is chip's int pin connect system's irq number.
 *                   param5 is object mode's device struct
 * Output        :   null
 * Return        :   ISIL_OK is meaning chip init ok, other is meaning chip init err.
 ****************************************************************************************************************/
void pci_init_ad(isil_chip_t *);
static int sys_clk = 0;
module_param(sys_clk, int, S_IRUGO);
MODULE_PARM_DESC(sys_clk, "On chip sys_clk rate, available sys_clk(108,110,120,150,166,172,175,180,185)");
static int  isil_chip_init(int bus_id, int chip_id, DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY *platform_encode_cap, int irq_id, struct device *dev)
{
    isil_chip_t	*chip;
    DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY	*video_encode_cap;
    isil_proc_register_s *chip_proc;
    int ret = 0;
    struct cmd_arg data;
    struct isil_chip_device *tcd;

    tcd = tc_chipdev_alloc(sizeof(*chip));
    if(!tcd){
        ret = -ENOMEM;
        return ret;
    }
    chip = tcd_priv(tcd);

    sprintf(chip->name, "isil_chip_%d_%d", bus_id, chip_id);
    chip->bus_id = bus_id;
    chip->chip_id = chip_id;

    if(sys_clk == 0) {
        chip->sys_clock = DEFAULT_SYS_CLK;
    }else{
        chip->sys_clock = sys_clk;
    }
    chip->dev = dev;
    atomic_set(&chip->open_chan_number, 0);
    init_robust_processing_control(&chip->chip_robust_processing);
    chip->video_encode_cap = platform_encode_cap;
    video_encode_cap = chip->video_encode_cap;
    chip->first_device_id = video_encode_cap->first_device_number;

    atomic_set(&chip->isil_int_control.chip_irq_register_flags, 0);
    chip->isil_int_control.chip_irq = irq_id;

    if(video_encode_cap->fpga_async_section_len != 0){
        chip->resource = request_mem_region(video_encode_cap->fpga_async_base_addr, video_encode_cap->fpga_async_section_len, chip->name);
        if(chip->resource == NULL){
            ISIL_DBG(ISIL_DBG_INFO, ": requeset mem regions err\n");
            goto release;
        }
        chip->regs = ioremap_nocache(video_encode_cap->fpga_async_base_addr, video_encode_cap->fpga_async_section_len);
        chip->pa_regs = video_encode_cap->fpga_async_base_addr;
    } else {
        chip->regs = NULL;
        chip->pa_regs = 0;
    }
    if(video_encode_cap->fpga_sync_section_len != 0){
        chip->sync_resource = request_mem_region(video_encode_cap->fpga_sync_base_addr, video_encode_cap->fpga_sync_section_len, chip->name);
        if(chip->sync_resource == NULL){
            ISIL_DBG(ISIL_DBG_INFO, ": requeset mem regions err\n");
            goto release;
        }
        chip->sync_regs = ioremap_nocache(video_encode_cap->fpga_sync_base_addr, video_encode_cap->fpga_sync_section_len);
        chip->pa_sync_regs = video_encode_cap->fpga_sync_base_addr;
    } else {
        chip->sync_regs = NULL;
        chip->pa_sync_regs = 0;
    }

    chip->chip_open = chip_hardware_resource_open;
    chip->chip_close = chip_hardware_resource_close;
    chip->chip_set_invalid = chip_phy_invalid;
    chip->chip_set_valid = chip_phy_valid;
    chip->chip_sysclk_init = chip_sysclk_init;
    chip->chip_ddr_init = chip_ddr_init;
    chip->chip_ddr_size_map_init = chip_ddr_size_map_init;
    chip->chip_auto_scan_ddr_size_map_mode = chip_auto_scan_ddr_size_map_mode;
    chip->chip_reset = chip_phy_reset;
#if defined(PLATFORM_ENDIAN_SAME)
    chip->io_op = &isil_host_io;
#elif defined(PLATFORM_ENDIAN_DIFFERENT)
    chip->io_op = &isil_pci_io;
#else
#warning "ENDIAN error!"
#endif

#if defined(PCI_PLATFORM)
    /*configure AD*/
    //pci_init_ad(chip);
#endif

    chip->register_timer   = register_on_chip_timer;
    chip->enable_timer     = enable_on_chip_timer;
    chip->disable_timer    = disable_on_chip_timer;
    chip->unregister_timer = unregister_on_chip_timer;
    chip->chip_set_invalid(chip);
    chip->chip_ddr_init(chip);
 
    chip_proc = &chip->chip_proc;
    chip_proc->entry = NULL;
    strcpy(chip_proc->name, "chip");
    chip_proc->read  = chip_proc_read;
#if defined(PCI_PLATFORM)
    chip_proc->write = chip_proc_write;
#else
    chip_proc->write = NULL;
#endif

    chip_proc->private = chip;
    isil_module_register(chip, chip_proc);

    init_chip_encode_chan_service_queue(&chip->chip_encode_service_queue);
    init_chip_ddr_burst_engine(&chip->chip_ddr_burst_interface, chip);
    init_chip_bus(&chip->isil_chip_bus, chip->bus_id, chip->chip_id);//never fail
    if(init_isil_interrupt_control(&chip->isil_int_control, chip) != ISIL_OK){
        ISIL_DBG(ISIL_DBG_FATAL, "irq controler init failed\n");
        goto remove;
    }
    chip->vlc_ping_pong_index = 0;
    chip->curr_h264_encode_chan = NULL;
    chip->register_curr_h264_encode_chan = register_curr_h264_encode_chan;
    chip->unregister_curr_h264_encode_chan = unregister_curr_h264_encode_chan;
    chip->register_curr_h264_encode_push_chan = register_curr_h264_encode_push_chan;
    chip->unregister_curr_h264_encode_push_chan = unregister_curr_h264_encode_push_chan;
    init_isil_timer(&chip->chip_timer_cont);

    if(tc_chip_init(tcd,bus_id,chip_id))
        goto remove;

    mutex_lock(&pool_lock);
    data.type = 5;//ISIL_AUDIO_IN
    data.channel_idx = 0;
    data.algorithm = 0;
    data.stream = 0;
    if(create_service_device(tcd, (void *)&data, 1)) {
	    mutex_unlock(&pool_lock);
	    goto err;
    }
    data.type = 3;//ISIL_VIDEO_IN
    if(create_service_device(tcd, (void *)&data, 1)) {
	    mutex_unlock(&pool_lock);
        goto err;
    }

#if defined(PCI_PLATFORM)
	if (push_buf_pool_init(chip)) {
		printk("failed to init pool!\n");
		mutex_unlock(&pool_lock);
		goto err;
	}
#endif

	chip->chip_set_valid(chip);
	mutex_unlock(&pool_lock);
	return ISIL_OK;    
err:
	ret = ISIL_ERR;
	release_all_service_device_ext_on_chip(tcd);
	remove_chip_device(tcd);
remove:
	isil_chip_remove(tcd);
release:
	release_chip_memory(tcd);
	ISIL_DBG(ISIL_DBG_INFO, ": %s_probe fail\n", chip->name);
	return ret;
}



#if defined(PCI_PLATFORM)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
/* __devinit and __devexit removed in these versions */
#define __devinit
#define __devexit
#define __devinit_p(x) x
#define __devexit_p(x) x
#endif

static int __devinit isil_pci_probe(struct pci_dev *pci_dev, const struct pci_device_id *id)
{
	int slot, bus_id, devfn, i;
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,11,0)
	struct pci_dev *bridge;
#endif
	struct pci_bus *bus;
	type_bus_t      *isil_pci_bus;
	isil_chip_bus_t *isil_chip_bus;
	isil_chip_t     *chip;
	u32     chip_id;
	int     ret = -1;
	static int max_dev = 0;
	DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY *platform_encode_cap;

	bus_id = pci_dev->bus->number;
	devfn  = PCI_FUNC(pci_dev->devfn);
	slot   = PCI_SLOT(pci_dev->devfn);

	bus = pci_dev->bus;
	printk("S819:: probe %d\n", ++max_dev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,11,0)
	// This code seems unnecessary and may cause lockups on 4.11
	while(bus && (bridge = bus->self)){
#ifndef BUILD_CENTOS7
		pci_enable_bridges(bus);
#endif	
		pci_set_master(pci_dev);
		if(pci_enable_device(bridge)){
			ISIL_DBG(ISIL_DBG_ERR, "enable pci device failed %d\n", ret);
		}
		ISIL_DBG(ISIL_DBG_INFO, "enable bridge %04x:%04x at %s\n", bridge->vendor, bridge->device, bus->name);
		bus = bridge->bus;
	}
#endif
	ret = pci_set_dma_mask(pci_dev, DMA_BIT_MASK(32));
	if(ret){
		ISIL_DBG(ISIL_DBG_ERR, "No suitable DMA available.\n");
		return ret;
	}
	ret = pci_set_consistent_dma_mask(pci_dev, DMA_BIT_MASK(32));
	if (ret) {
		ISIL_DBG(ISIL_DBG_ERR, "No usable consistent DMA configuration, aborting\n");
		return ret;
	}
	ret = pci_enable_device(pci_dev);
	if(ret){
		ISIL_DBG(ISIL_DBG_ERR, "enable pci device failed %d\n", ret);
		return ret;
	}

	ISIL_DBG(ISIL_DBG_INFO, "detect tw%x pci device, bus %d, slot %d, fn %d, irq %d\n", pci_dev->device, bus_id, slot, devfn, pci_dev->irq);
	for(i = 0; i < DEVICE_COUNT_RESOURCE; i++)
	{
		if(pci_dev->resource[i].start){
			ISIL_DBG(ISIL_DBG_INFO, "BAR%d  %s Addr 0x%lx, Size 0x%lx\n", 
				 i, (pci_dev->resource[i].flags & IORESOURCE_IO) ? "I/O":"MEM",
				 (unsigned long)(pci_resource_start(pci_dev, i)),
				 (unsigned long)(pci_resource_len(pci_dev, i)));
		}
	}

	get_isil_pci_bus(&isil_pci_bus);
	chip_id = isil_pci_bus->op->get_device_id(isil_pci_bus);
	platform_encode_cap = (DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY *)kzalloc(sizeof(DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY), GFP_KERNEL);
	if(!platform_encode_cap) {
		return -ENOMEM;
	}
	memcpy(platform_encode_cap, &platform_video_encode_cap[0], sizeof(DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY));
	platform_encode_cap->first_device_number = chip_id * ISIL_PHY_VD_CHAN_NUMBER;
	platform_encode_cap->fpga_async_base_addr =  pci_resource_start(pci_dev, 0);//TW5864 only use BAR0
	platform_encode_cap->fpga_async_section_len = pci_resource_len(pci_dev, 0);

	ret = isil_chip_init(ISIL_ED_PCI_BUS, chip_id, platform_encode_cap, pci_dev->irq, &pci_dev->dev);
	if(ret){
		ISIL_DBG(ISIL_DBG_FATAL, "init chip failed %d\n", ret);
		kfree(platform_encode_cap);
		pci_disable_device(pci_dev);
		return ret;
	}

	get_isil_pci_bus(&isil_pci_bus);
	isil_pci_bus->op->find_chip_bus_in_type_bus_table(isil_pci_bus, &isil_chip_bus, chip_id);
	if(isil_chip_bus != NULL){
		chip = to_get_chip_with_chip_bus(isil_chip_bus);
		/*enable bus control*/
		pci_set_master(pci_dev);
		/*adjust latency timer value*/
		//pci_write_config_dword(pci_dev, 0xc, 0x8008);
		pci_set_drvdata(pci_dev, chip);
	}else{
		ISIL_DBG(ISIL_DBG_ERR, "No such chip %d\n", chip_id);
		pci_disable_device(pci_dev);
	}
	return 0;
} 

static void __devexit isil_pci_remove(struct pci_dev *pci_dev)
{
    type_bus_t      *isil_pci_bus;
    isil_chip_bus_t *isil_chip_bus;
    isil_chip_t     *chip;
    struct isil_chip_device *tcd;
    DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY *video_encode_cap;

    chip = pci_get_drvdata(pci_dev);
    get_isil_pci_bus(&isil_pci_bus); 
    isil_pci_bus->op->find_chip_bus_in_type_bus_table(isil_pci_bus, &isil_chip_bus, chip->chip_id);
    if(isil_chip_bus != NULL){
        tcd = priv_2_tcd((void*)chip);
        release_all_service_device_ext_on_chip(tcd);
        remove_chip_device(tcd);
        video_encode_cap = chip->video_encode_cap;
        isil_chip_remove(tcd);
        release_chip_memory(tcd);
        if(video_encode_cap) {
            kfree(video_encode_cap);
        }
    }else{
        ISIL_DBG(ISIL_DBG_INFO, ":  chip not register!!\n");
    }
    pci_disable_device(pci_dev);
}

static struct pci_device_id isil_pci_tbl[]  = {
    {
        .vendor = PCI_ISIL_VENDOR_ID,
        .device = PCI_DEVICE_ID_TW5864,
        .subvendor = PCI_ANY_ID, 
        .subdevice = PCI_ANY_ID, 
        .class = 0,
        .class_mask = 0,
        .driver_data = 0,
    },
    {
        .vendor = PCI_ISIL_VENDOR_ID,
        .device = PCI_DEVICE_ID_TW5866,
        .subvendor = PCI_ANY_ID, 
        .subdevice = PCI_ANY_ID, 
        .class = 0,
        .class_mask = 0,
        .driver_data = 0,
    },{},
};

static struct pci_driver isil_pci_driver = {
    .name     = "TW5864",
    .id_table = isil_pci_tbl,
    .probe    = isil_pci_probe,
    .remove   = isil_pci_remove,
};
#else
static int	async_bus_chip_driver_init(void)
{
    type_bus_t      *isil_host_bus;
    DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY *platform_encode_cap;
    u32	chip_id;
    int	ret;

    get_isil_host_bus(&isil_host_bus);
    chip_id = isil_host_bus->op->get_device_id(isil_host_bus);
    platform_encode_cap = (DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY *)kzalloc(sizeof(DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY), GFP_KERNEL);
    if(!platform_encode_cap) {
        return -ENOMEM;
    }
    memcpy(platform_encode_cap, &platform_video_encode_cap[0], sizeof(DVMNVS_CHIP_VIDEO_ENCODE_CAPABILITY));
    platform_encode_cap->first_device_number = chip_id * ISIL_PHY_VD_CHAN_NUMBER;
    ret = isil_chip_init(ISIL_ED_HOST_BUS, chip_id, platform_encode_cap, IRQ_EINT1, NULL);
    if(ret){
        ISIL_DBG(ISIL_DBG_FATAL, "init chip failed %d\n", ret);
        kfree(platform_encode_cap);
        return ret;
    }
    return 0;
}
#endif

/****************************************************************************************************************
 * Function Name :   isil_init
 * Description   :   tw5864 module insmod's entrance chip's.
 * Input         :   null
 * Output        :   null
 * Return        :   ISIL_OK is meaning insmod ok, other is insmod err.
 ****************************************************************************************************************/
static int __init   isil_init(void)
{
    int ret = ISIL_OK;

    printk("\nVersion: %d.%d.%d-%d\nGCC Version: %s\nARCH: %s\nBuild at: __DATE__ __TIME__\n", 
            get_version_major(__ISIL__CODE__VERSION__),
            get_version_submajor(__ISIL__CODE__VERSION__),
            get_version_minor(__ISIL__CODE__VERSION__), 
            __ISIL_SVN_VERSION__,__VERSION__, __ISIL_ARCH__ );

    init_system_walltimer();
    isil_core_init();
    register_all_tc_device();
    register_all_tc_driver();
    init_isil_dma_chan();
    ret = isil_chip_driver_init();
    if(ret != ISIL_OK){
        ISIL_DBG(ISIL_DBG_INFO, ": chip driver register failed!\n");
        return ret;
    }

    proc_module_init();
    init_isil_root_bus();
#ifdef  POWERPC_PLATFORM
    isil_map_init();
#endif
    mutex_init(&pool_lock);
#if defined(PCI_PLATFORM)
    if (pci_register_driver(&isil_pci_driver)) {
#if defined(POWERPC_PLATFORM)
        isil_map_exit();
#endif
        remove_isil_root_bus();
        ISIL_DBG(ISIL_DBG_INFO, ": Faild register pci driver!\n");
        return -ENODEV;
    }
#else
    async_bus_chip_driver_init();
#endif
    return ret;
}

static void __exit  isil_exit(void)
{
#if  defined(ARM_PLATFORM)
    struct isil_chip_device *tcd;
    type_bus_t            *isil_host_bus;
    isil_chip_bus_t       *isil_chip_bus;
    isil_chip_t *chip;
#endif
#if  defined(PCI_PLATFORM)
    pci_unregister_driver(&isil_pci_driver);
#else
    get_isil_host_bus(&isil_host_bus);
    isil_host_bus->op->find_chip_bus_in_type_bus_table(isil_host_bus, &isil_chip_bus, 0);
    if(isil_chip_bus != NULL){
        chip = to_get_chip_with_chip_bus(isil_chip_bus);

        tcd = priv_2_tcd((void*)chip);
        release_all_service_device_ext_on_chip(tcd);
        remove_chip_device(tcd);
        isil_chip_remove(tcd);
        release_chip_memory(tcd);
    }
#endif

    remove_isil_root_bus();
    isil_chip_driver_remove();
    unregister_all_tc_driver();
    unregister_all_tc_devices();
    proc_module_release();
    isil_core_exit();
#ifdef  POWERPC_PLATFORM
    isil_map_exit();
#endif
}

MODULE_LICENSE("Dual BSD/GPL");

module_init(isil_init);
module_exit(isil_exit);

